"""The convert module contains transformations related to a conversion format."""
from pygame import transform as tf, Surface
from .transformation import Transformation

class GrayScale(Transformation):
    """
    The gray scale transformation turns the art into a black and white art. The frames are converted in a 8-bits-per-pixel format.
    """

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        graysurfeaces = tuple(tf.grayscale(surf) for surf in surfaces)
        return graysurfeaces, durations, introduction, None, width, height

class ConvertRGBA(Transformation):
    """The convert RGBA tranformation adds an alpha layer to the art. The frames are converted in a 32-bits-per-pixel format."""

    def apply(self, surfaces, durations, introduction, index, width, height, **ld_kwargs):
        surfaces = tuple(surf.convert_alpha() for surf in surfaces)
        return surfaces, durations, introduction, None, width, height

class ConvertRGB(Transformation):
    """The convert RGB transformation converts the frames in a 24-bits-per-pixel format. The alpha layer is removed."""
    def apply(self, surfaces, durations, introduction, index, width, height, **ld_kwargs):
        surfaces = tuple(surf.convert() for surf in surfaces)
        return surfaces, durations, introduction, None, width, height



"""The geometry module contains arts build from a geometry."""
from typing import Sequence
from pygame import Surface, SRCALPHA, mask as msk, Color, gfxdraw, draw
from pygamecv import rectangle, circle, ellipse, polygon, rounded_rectangle
from .art import Art
from ..transform import Transformation
from .._common import ColorValue

class Rectangle(Art):
    """A Rectangle is an Art representing a rectangle."""

    def __init__(
        self,
        color: ColorValue,
        width: int,
        height: int,
        thickness: int = 0,
        transformation: Transformation = None,
    ):
        """
        A Rectangle is an Art representing a rectangle.

        Params:
        ---
        - color: ColorValue, the color of the rectangle.
        - width: int, the width of the art.
        - height: int, the height of the art. 
        - thickness: int, the thickness of the line used to draw the art.
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        """

        super().__init__(transformation)
        self.color = Color(color)
        self._initial_width, self.initial_height = width, height
        self._width = width
        self._height = height
        self.thickness = thickness
        self._find_initial_dimension()

    def _load(self, **ld_kwargs):
        surf = Surface((self._initial_width, self.initial_height), SRCALPHA if self.color.a != 255 and self.thickness != 0 else 0)
        # Pygamecv's paradigm is: defining the control points and let the thickness go outside of the geometry
        # Here, we unsure the width and height are excatly what is asked to be, not width + thickness and height + thickness
        rect = (self.thickness//2, self.thickness//2, self._initial_width - self.thickness, self.initial_height - self.thickness)
        rectangle(surf, rect, self.color, self.thickness)
        self.surfaces = (surf,)
        self._durations = (0,)

class RoundedRectangle(Art):
    """A RoundedRectangle is an Art representing rounded rectangle."""

    def __init__(
        self,
        color: ColorValue,
        width: int,
        height: int,
        top_left: int,
        top_right: int = None,
        bottom_left: int = None,
        bottom_right: int = None,
        thickness: int = 0,
        transformation: Transformation = None,
        allow_antialias: bool = True,
        background_color: ColorValue = None
    ):
        """
        A RoundedRectangle is an Art representing rounded rectangle.

        Params:
        ---
        - color: ColorValue, the color of the rounded rectangle.
        - width: int, the width of the art.
        - height: int, the height of the art. 
        - top_left, top_right, bottom_left, bottom_right: the radii of the rounded corners.
        If one of the 3 last is None, the value for the top_left corner is used instead.
        - thickness: int, the thickness of the line used to draw the art.
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        - background_color: ColorValue, the color of the background. This is used to make better renders when antialias is used.
        """
        super().__init__(transformation)
        self.top_left = top_left
        self.top_right = top_right
        self.bottom_left = bottom_left
        self.bottom_right = bottom_right
        self.color = Color(color)
        self.thickness = thickness
        self._width = self._initial_width = width
        self._height = self.initial_height = height
        self.allow_antialias = allow_antialias
        self.background_color = background_color
        self._find_initial_dimension()

    def _load(self, **ld_kwargs):
        surf = Surface((self._initial_width, self.initial_height), SRCALPHA)
        if self.background_color is not None: # Useful in case of antialias
            surf.fill((*self.background_color[:3], 0))
        rect = (self.thickness//2, self.thickness//2, self._initial_width - self.thickness, self.initial_height - self.thickness)
        rounded_rectangle(surf, rect, self.color, self.thickness, ld_kwargs.get("antialias", False) and self.allow_antialias,
                          self.top_left, self.top_right, self.bottom_left, self.bottom_right)
        self._surfaces = (surf,)
        self._durations = (0,)

class Circle(Art):
    """A Circle is an Art representing a circle."""

    def __init__(
        self,
        color: ColorValue,
        radius: int,
        thickness: int = 0,
        transformation: Transformation = None,
        allow_antialias: bool = True,
        background_color: Color = None
    ):
        """
        A Circle is an Art representing a circle.

        Params:
        ---
        - color: ColorValue, the color of the circle.
        - radius: int, the radius of circle. The art will have a width and height of twice the radius.
        - thickness: int, the thickness of the line used to draw the art.
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        - background_color: ColorValue, the color of the background. This is used to make better renders when antialias is used.
        """
        super().__init__(transformation)
        self.radius = radius
        self.color = Color(color)
        self.thickness = thickness
        self._height = 2*radius
        self._width = 2*radius
        self.allow_antialias = allow_antialias
        self.background_color = background_color
        self._find_initial_dimension()

    def _load(self, **ld_kwargs):
        surf = Surface((self.radius*2, self.radius*2), SRCALPHA)
        if self.background_color is not None: # Useful in case of antialias
            surf.fill((*self.background_color[:3], 0))
        radius = self.radius - self.thickness//2
        if radius > 0:
            circle(surf, (self.radius, self.radius), radius, self.color,
                self.thickness, ld_kwargs.get("antialias", False) and self.allow_antialias)
        else:
            circle(surf, (self.radius, self.radius), self.radius, self.color,
                0, ld_kwargs.get("antialias", False) and self.allow_antialias)

        self._surfaces = (surf,)
        self._durations = (0,)

class Ellipse(Art):
    """An Ellipse is an Art with a colored ellipse at the center."""

    def __init__(
        self,
        color: Color,
        radius_x: int,
        radius_y: int,
        thickness: int = 0,
        transformation: Transformation = None,
        allow_antialias: bool = True,
        background_color: Color = None
    ) -> None:
        """
        An Ellipse is an Art representing a ellipse.

        Params:
        ---
        - color: ColorValue, the color of the the ellipse.
        - radius_x: int, the horizontal radius of the ellipse. The width of the art is of twice this value. 
        - radius_y: int, the vertical radius of the ellipse. The height of the art is of twice this value. 
        - thickness: int, the thickness of the line used to draw the art.
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        - background_color: ColorValue, the color of the background. This is used to make better renders when antialias is used.
        """
        self.color = color
        self.thickness = thickness
        super().__init__(transformation)
        self.radius_x, self.radius_y = radius_x, radius_y
        self._height = radius_y*2
        self._width = radius_x*2
        self.allow_antialias = allow_antialias
        self.background_color = background_color
        self._find_initial_dimension()

    def _load(self, **ld_kwargs):
        surf = Surface((self.radius_x*2, self.radius_y*2), SRCALPHA)
        if self.background_color is not None: # Useful in case of antialias
            surf.fill((*self.background_color[:3], 0))
        radius_x = self.radius_x - self.thickness //2
        radius_y = self.radius_y - self.thickness //2
        if radius_x > 0 and radius_y > 0:
            ellipse(surf, (self.radius_x, self.radius_y), radius_x, radius_y, self.color, self.thickness, ld_kwargs.get("antialias", False) and self.allow_antialias, 0)
        else:
            ellipse(surf, (self.radius_x, self.radius_y), self.radius_x, self.radius_y, self.color, 0, ld_kwargs.get("antialias", False) and self.allow_antialias, 0)
        self._surfaces = (surf,)
        self._durations = (0,)

class Polygon(Art):
    """A Polygon is an Art representing a polygon."""

    def __init__(
        self,
        color: Color,
        points: Sequence[tuple[int, int]],
        thickness: int = 0,
        transformation: Transformation = None,
        allow_antialias: bool = True,
        background_color: Color = None
    ):
        """
        A Circle is an Art representing a polygon.

        Params:
        ---
        - color: ColorValue, the color of the polygon.
        - points: Sequence[tuple[int, int]], the points used to draw the polygon. All points will be translated in the plan so that
        min(p[0] for p in points) == min(p[1] for p in points) == 0. The width and height of the art are equal to
        max(p[0] for p in points) and max(p[1] for p in points) after the translation.
        - thickness: int, the thickness of the line used to draw the art.
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        - background_color: ColorValue, the color of the background. This is used to make better renders when antialias is used.
        """

        self.points = points
        self.thickness = thickness
        self.color = color
        super().__init__(transformation)

        min_x = min(p[0] for p in self.points)
        min_y = min(p[1] for p in self.points)

        self.points = [(p[0] - min_x + thickness//2, p[1] - min_y + thickness//2) for p in points]

        self._height = self._initial_height = max(p[1] for p in self.points) + thickness//2
        self._width = self._initial_width = max(p[0] for p in self.points) + thickness//2
        self.allow_antialias = allow_antialias
        self.background_color = background_color
        self._find_initial_dimension()

    def _load(self, **ld_kwargs):

        surf = Surface((self._initial_width, self._initial_height), SRCALPHA)
        if self.background_color is not None:
            surf.fill((*self.background_color[:3], 0))
        polygon(surf, self.points, self.color, self.thickness, ld_kwargs.get("antialias", False) and self.allow_antialias)

        self._surfaces = (surf,)
        self._durations = (0,)

class TexturedPolygon(Art):
    """A Textured polygon represents a polygon filled with an art.."""

    def __init__(
        self,
        texture: Art,
        points: Sequence[tuple[int, int]],
        transformation: Transformation = None,
    ):
        """
        A Textured polygon represents a polygon filled with an art.
        
        Params:
        ----
        - texture: Art, the Art drawn inside the polygon. The TexturedPolygon will have the same width, height, durations and introduction
        as its texture. The surfaces use to create the TexturedPolygon are the surfaces of the texture when the TexturedPolygon is loaded.
        - points: Sequence[tuple[int, int]] the list of points used to draw the polygon.
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        """

        self.points = points
        super().__init__(transformation)
        self._height = max(p[1] for p in self.points)
        self._width = max(p[0] for p in self.points)
        self._find_initial_dimension()
        self.texture = texture

    def _load(self, **ld_kwargs):

        surfaces = []
        need_to_unload = False

        if not self.texture.is_loaded:
            need_to_unload = True
            self.texture.load(**ld_kwargs)

        else: # the texture might have change, so can its dimensions.
            self._width = self.texture.width
            self._height = self.texture.height
            self._find_initial_dimension()

        for surf in self.texture.surfaces:
            background = Surface((self._width, self._height), SRCALPHA)
            gfxdraw.textured_polygon(background, self.points, surf.convert_alpha(), 0, 0)
            surfaces.append(background)

        self._surfaces = tuple(surfaces)
        self._durations = self.texture.durations
        self._introduction = self.texture.introduction

        if need_to_unload:
            self.texture.unload()

class TexturedCircle(Art):
    """A TexturedCircle is an Art with a textured circle at the center of it."""

    def __init__(
        self,
        texture: Art,
        radius: int,
        center: tuple[int, int] = None,
        draw_top_right: bool = True,
        draw_top_left: bool = True,
        draw_bottom_left: bool = True,
        draw_bottom_right: bool = True,
        transformation: Transformation = None,
    ):
        """
        A TexturedCircle represents a circle filled with an art.
        
        Params:
        ----
        - texture: Art, the Art drawn inside the polygon. The TexturedPolygon will have the same width, height, durations and introduction
        as its texture. The surfaces use to create the TexturedPolygon are the surfaces of the texture when the TexturedPolygon is loaded.
        - radius: the radius of the circle
        - center: the center of the circle. By default, the center of the Art. The circle may not be fully drawn.
        - draw_top_right, draw_top_left, draw_bottom_left, draw_bottom_right: bool, specify whether the corresponding quart should be drawn.
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        """

        super().__init__(transformation)
        self.radius = radius
        self.draw_top_right = draw_top_right
        self.draw_top_left = draw_top_left
        self.draw_bottom_left = draw_bottom_left
        self.draw_bottom_right = draw_bottom_right
        if center is None:
            center = texture.width//2, texture.height//2
        self.center = center
        self._width = texture.width
        self._height = texture.height
        self.texture = texture
        self._find_initial_dimension()

    def _load(self, **ld_kwargs):

        need_to_unload = False
        if not self.texture.is_loaded:
            need_to_unload = True
            self.texture.load(**ld_kwargs)

        else: # the texture might have change, so can its dimensions.
            self._width = self.texture.width
            self._height = self.texture.height
            self._find_initial_dimension()

        surf = Surface((self._width, self._height), SRCALPHA)
        draw.circle(surf, (255, 255, 255, 255), self.center,
            self.radius, 0, self.draw_top_right, self.draw_top_left, self.draw_bottom_left, self.draw_bottom_right)
        mask = msk.from_surface(surf, 127)
        self._surfaces = tuple(mask.to_surface(setsurface=surface.convert_alpha(), unsetsurface=surf) for surface in self.texture.surfaces)
        self._durations = self.texture.durations
        self._introduction = self.texture.introduction


        if need_to_unload:
            self.texture.unload()

class TexturedEllipse(Art):
    """A TexturedEllipse represents an ellipse filled with an art."""

    def __init__(
        self,
        texture: Art,
        radius_x: int,
        radius_y: int,
        center: tuple[int, int] = None,
        transformation: Transformation = None,
    ) -> None:
        """
        A TexturedEllipse represents an ellipse filled with an art.
        
        Params:
        ----
        - texture: Art, the Art drawn inside the polygon. The TexturedPolygon will have the same width, height, durations and introduction
        as its texture. The surfaces use to create the TexturedPolygon are the surfaces of the texture when the TexturedPolygon is loaded.
        - radius_x, radius_y: int, the horizontal and vertical radii of the ellipse.
        - center: tuple[int, int], default is the center of the art. The center of the ellipse.
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        """
        super().__init__(transformation)
        if center is None:
            center = texture.width//2, texture.height//2
        self.center = center
        self.rect = (self.center[0] - radius_x, self.center[0] - radius_y, radius_x*2, radius_y*2)
        self._width = texture.width
        self._height = texture.height
        self.texture = texture
        self._find_initial_dimension()

    def _load(self, **ld_kwargs):

        need_to_unload = False
        if not self.texture.is_loaded:
            need_to_unload = True
            self.texture.load(**ld_kwargs)
        
        else: # the texture might have change, so can its dimensions.
            self._width = self.texture.width
            self._height = self.texture.height
            self._find_initial_dimension()

        surf = Surface((self._width, self._height), SRCALPHA)
        draw.ellipse(surf, (255, 255, 255, 255), self.rect, 0)
        mask = msk.from_surface(surf, 127)
        self._surfaces = tuple(mask.to_surface(setsurface=surface.convert_alpha(), unsetsurface=surf) for surface in self.texture.surfaces)
        self._durations = self.texture.durations
        self._introduction = self.texture.introduction

        if need_to_unload:
            self.texture.unload()

class TexturedRoundedRectangle(Art):
    """A TexturedRoundedRectangle is an Art with rounded corners."""

    def __init__(
        self,
        texture: Art,
        top_left: int,
        top_right: int = None,
        bottom_left: int = None,
        bottom_right: int = None,
        transformation: Transformation = None,
    ):
        """
        A TexturedRoundedRectangle is an Art with rounded corners.
        
        Params:
        ----
        - texture: Art, the Art drawn inside the polygon. The TexturedPolygon will have the same width, height, durations and introduction
        as its texture. The surfaces use to create the TexturedPolygon are the surfaces of the texture when the TexturedPolygon is loaded.
        - top_left, top_right, bottom_left, bottom_right: int. The radii of the corners. If any of the 3 last is None, it it repplaced by the
        value for top_left.
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        """

        super().__init__(transformation)
        self.top_left = top_left
        self.top_right = top_right if not top_right is None else top_left
        self.bottom_left = bottom_left if not bottom_left is None else top_left
        self.bottom_right = bottom_right if not bottom_right is None else top_left
        self._height = texture.height
        self._width = texture.width
        self.texture = texture

        self._find_initial_dimension()

    def _load(self, **ld_kwargs):

        need_to_unload = False
        if not self.texture.is_loaded:
            need_to_unload = True
            self.texture.load(**ld_kwargs)

        surf = Surface((self.width, self.height), SRCALPHA)
        draw.rect(
            surf,
            (255, 255, 255, 255),
            (0, 0, self.width, self.height),
            0,
            -1,
            self.top_left,
            self.top_right,
            self.bottom_left,
            self.bottom_right
        )
        mask = msk.from_surface(surf, 127)
        self._surfaces = tuple(mask.to_surface(setsurface=surface.convert_alpha(), unsetsurface=surf) for surface in self.texture.surfaces)
        self._durations = self.texture.durations
        self._introduction = self.texture.introduction

        if need_to_unload:
            self.texture.unload()



"""The transformation module contains the base class Transformation and all the subclasses."""
from typing import Iterable
from itertools import cycle
from abc import ABC, abstractmethod
from math import cos, sin, radians
from random import randint, shuffle
import pygame.transform as tf
from pygame import Surface, SRCALPHA, Rect
from .._common import ColorValue

class Transformation(ABC):
    """
    A transformation is an operation on an art. This class is an abstract class and shouldn't be instanciated.
    """

    @abstractmethod
    def apply(
        self,
        surfaces: tuple[Surface],
        durations: tuple[int],
        introduction: int,
        index: int,
        width: int,
        height: int,
        **ld_kwargs
    ):
        """Apply the transformation"""
        raise NotImplementedError()

    def get_new_dimension(self, width, height) -> tuple[int, int]:
        """Calculate the new dimensions of the art after transformation."""
        return width, height

    # pylint: disable=unused-argument
    def cost(self, width, height, length, **ld_kwargs):
        """
        Return the cost of the transformation.
        
        The cost is a numerical metric used to determine whether the transformation
        should be done in the main thread or in a parallel thread.
        """
        return 0

    def __len__(self):
        return 1

class Pipeline(Transformation):
    """A Transformation pipeline is a sequence of successive transformations."""

    def __init__(self, *transfos) -> None:
        """
        A Pipeline is a sequence of successive transformation.
        
        Params:
        ---
        - *transfos: the succesive transformations.
        """
        super().__init__()
        self._transformations: list[Transformation] = list(transfos).copy()

    def add_transformation(self, transfo: Transformation) -> None:
        """Add a new transformation in the pipeline."""
        self._transformations.append(transfo)

    def clear(self):
        """Clear the Pipeline from all transformations."""
        self._transformations.clear()
    
    def __len__(self):
        return sum(len(transfo) for transfo in self._transformations)

    def is_empty(self) -> bool:
        """Return True if the Pipeline is empty of transformations."""
        return not self._transformations

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        successive_indices = []
        for transfo in self._transformations:
            surfaces, durations, introduction, idx, width, height = transfo.apply(surfaces, durations, introduction, index, width, height, **ld_kwargs)
            if idx is not None:
                index = idx
            successive_indices.append(idx)
        # Find the last not None index. If there is no, we return None
        for idx in successive_indices[::-1]:
            if idx is not None:
                break
        return surfaces, durations, introduction, idx, width, height

    def get_new_dimension(self, width, height):
        for transfo in self._transformations:
            width, height = transfo.get_new_dimension(width, height)
        return width, height

    def cost(self, width, height, length, **ld_kwargs):
        return sum(transfo.cost(width, height, length, **ld_kwargs) for transfo in self._transformations)

    def copy(self):
        """
        Return a copy of the Pipeline.
        """
        return Pipeline(*self._transformations)

class Rotate(Transformation):
    """The Rotate transformation rotates the art by a given angle."""

    def __init__(self, angle: float) -> None:
        """
        The Rotate transformation rotates the art by a given angle.
        
        Params:
        ---
        - angle: float, the angle to counterclockwise rotate the art, in degrees.
        """
        super().__init__()
        self.angle = angle

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        rotated_surfaces = tuple(tf.rotate(surf, self.angle) for surf in surfaces)
        return rotated_surfaces, durations, introduction, None, *rotated_surfaces[0].get_size()

    def get_new_dimension(self, width, height):
        radians_angle = radians(self.angle)
        new_width = abs(width * cos(radians_angle)) + abs(height * sin(radians_angle))
        new_height = abs(width * sin(radians_angle)) + abs(height * cos(radians_angle))
        return int(new_width), int(new_height)


class Zoom(Transformation):
    """
    The Zoom transformation zoomes the art by a give scale.

    Example:
    ----
    If the art have a size of (100, 100), calling this transformation with a scale of 1.2 would modify the art
    to a size (120, 120). Calling this transformation with a scale of 0.6 would modify the art
    to a size (60, 60). You can also specify two scales (one for horizontal and one for vertical) by passing
    a tuple as scale. If smooth is True, use a smooth zooming instead.
    """

    def __init__(self, scale: float | tuple[float, float], smooth: bool = False) -> None:
        """
        The Zoom transformation zoomes the art by a give scale.

        Params:
        ---
        - scale: float | tuple[float, float]. If two numbers are provided, the scaling on each axis is done separately. The width and the height of the
        art are multiplied by the scale.
        - smooth: bool, whether smooth scaling should be used or not. Smooth scaling usually looks better but is more costly.
        """
        super().__init__()
        self.scale = scale
        self.smooth = smooth

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        if (self.scale == (2, 2) or self.scale == 2) and not self.smooth:
            rescaled_surfaces = tuple(tf.scale2x(surf) for surf in surfaces)
        elif not self.smooth:
            rescaled_surfaces = tuple(tf.scale_by(surf, self.scale) for surf in surfaces)
        else:
            rescaled_surfaces = tuple(tf.smoothscale_by(surf, self.scale) for surf in surfaces)
        return rescaled_surfaces, durations, introduction, None, int(width*self.scale), int(height*self.scale)

    def get_new_dimension(self, width, height):
        return int(width*self.scale), int(height*self.scale)

class Resize(Transformation):
    """
    The Resize transformation resizes the art to a new size. The image might end distorded.

    Example:
    ----
    If the art have a size of (100, 100), calling this transformation with a zie of (120, 60) would modify the art
    to a size (120, 60). If smooth is True, use a smooth resizing instead.
    """

    def __init__(self, size: tuple[int, int], smooth: bool = False) -> None:
        """
        The Resize transformation resizes the art to a new size. The image might end distorded.

        Params:
        ---
        - size: tuple[int, int], the new size of the images.
        - smooth: bool, whether smooth scaling should be used or not. Smooth scaling usually looks better but is more costly.
        """
        super().__init__()
        self.size = size
        self.smooth = smooth

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        if self.smooth:
            rescaled_surfaces = tuple(tf.smoothscale(surf, self.size) for surf in surfaces)
        else:
            rescaled_surfaces = tuple(tf.scale(surf, self.size) for surf in surfaces)
        return rescaled_surfaces, durations, introduction, None, *self.size

    def get_new_dimension(self, width, height):
        return self.size

class Crop(Transformation):
    """
    The Crop transformation crops the art to a smaller art.

    Example:
    ----
    If the art have a size of (100, 100), calling this transformation with left=50, top=50, width=20, height=30 will result
    in a surface with only the pixels from (50, 50) to (70, 80)
    """

    def __init__(self, left: int, top: int, width: int, height: int) -> None:
        """
        The Crop transformation crops the art to a smaller art.
        
        Params:
        ----
        - left, top: int, the coordinate of the top_left of the cropped area inside the full art.
        - width, height: int, the size of the cropped area.

        Note:
        ----
        - If left < 0, top < 0, width > art.width - left or height > art.height - top, the cropped surface is restricted to the art. The
        width and height of the final art might be different than the constructor's argument in this case.
        """

        super().__init__()
        self.rect = Rect(left, top, width, height)

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        background = Surface(self.rect.size)
        cropped_surfaces = []
        rect = self.rect.clip(surfaces[0].get_rect())
        for surf in surfaces:
            background.blit(surf, (0,0), rect)
            cropped_surfaces.append(background)
        return tuple(cropped_surfaces), durations, introduction, None, *rect.size

    def get_new_dimension(self, width, height):
        rect = self.rect.clip((0, 0, width, height))
        return rect.size

class Pad(Transformation):
    """
    The Pad transformation adds a solid color extension on every side of the art
    """

    def __init__(self, color: ColorValue, left: int = 0, right = 0, top = 0, bottom = 0) -> None:
        """
        The Pad transformation adds a solid color extension on every side of the art.

        Params:
        ----
        - color: ColorValue: the color used to pad the art.
        - left, right, top, bottom: int, the number of pixel to pad in each direction. If negative, acts like a crop.
        """
        super().__init__()
        self.color = color
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        background = Surface((width + self.left + self.right, height + self.left + self.right), SRCALPHA)
        background.fill(self.color)
        padded_surfaces = []
        for surf in surfaces:
            background.blit(surf, (self.left, self.top))
            padded_surfaces.append(background.copy())
        return tuple(padded_surfaces), durations, introduction, None, width + self.left + self.right, height + self.top + self.bottom

    def get_new_dimension(self, width, height):
        return width + self.left + self.right, height + self.top + self.bottom

class Flip(Transformation):
    """
    The flip transformation flips the art, horizontally and/or vertically.
    """

    def __init__(self, horizontal: bool, vertical: bool) -> None:
        """
        The flip transformation flips the art, horizontally and/or vertically.
        
        Params:
        ----
        - horizontal: bool, whether the flip should be done horizontally.
        - vercital: bool, whether the flip should be done vertically.
        """
        super().__init__()
        self.horizontal = horizontal
        self.vertical = vertical

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        flipped_surfaces = tuple(tf.flip(surf, self.horizontal, self.vertical) for surf in surfaces)
        return flipped_surfaces, durations, introduction, None, height, width

    def get_new_dimension(self, width, height):
        return width, height

class Transpose(Transformation):
    """The transpose transformation transposes the art like a matrix."""

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        tp_surfaces = tuple(tf.flip(tf.rotate(surf, 270), True, False) for surf in surfaces)
        return tp_surfaces, durations, introduction, None, width, height

    def get_new_dimension(self, width, height):
        return height, width

class VerticalChop(Transformation):
    """
    The vertical chop transformation removes a band of pixel and puts the right side next to the left side.
    """

    def __init__(self, from_: int, to: int) -> None:
        """
        The vertical chop transformation removes a band of pixel and puts the right side next to the left side.

        Params:
        ----
        - from_: int, the index of the first column of pixel to be removed.
        - to: int, the index of the last column of pixel to be removed.
        """
        super().__init__()
        self.rect = (from_, 0, to - from_ + 1, 0)

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        chopped_surfaces = tuple(tf.chop(surf, self.rect) for surf in surfaces)
        return chopped_surfaces, durations, introduction, None, width - self.rect[2], height

    def get_new_dimension(self, width, height):
        return width - self.rect[2], height

class HorizontalChop(Transformation):
    """
    The horizontal chop transformation removes a band of pixel and puts the bottom side just below to the top side.
    """

    def __init__(self, from_: int, to: int) -> None:
        """
        The horizontal chop transformation removes a band of pixel and puts the bottom side just below to the top side.

        Params:
        ----
        - from_: int, the index of the first row of pixel to be removed.
        - to: int, the index of the last row of pixel to be removed.
        """
        super().__init__()
        self.rect = (0, from_, 0, to - from_)

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        chopped_surfaces = tuple(tf.chop(surf, self.rect) for surf in surfaces)
        return chopped_surfaces, durations, introduction, None, width, height - self.rect[3]

    def get_new_dimension(self, width, height):
        return width, height - self.rect[3]

class SpeedUp(Transformation):
    """
    Speed up the animation by a scale.

    Example.
    If the duration of each frame in the art is 100 ms and the scale is 2, each frame lasts now 50 ms.
    """

    def __init__(self, scale: float) -> None:
        """
        Speed up the animation by a scale.

        Params:
        ----
        - scale: float > 0, the value by which each frame duration will be divided.

        Raises:
        ----
        - ValueError if the provided scale is null or negative.
        """
        super().__init__()
        if scale <= 0:
            raise ValueError(f"The scale cannot be 0 nor negative, got {scale}.")
        self.scale = scale

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        new_durations = tuple(d/self.scale for d in durations)
        return surfaces, new_durations, introduction, None, width, height

class SlowDown(Transformation):
    """
    Slow down the animation by a scale.

    Example.
    If the duration of each frame in the art is 100 ms and the scale is 2, each frame lasts now 200 ms.
    """

    def __init__(self, scale: float) -> None:
        """
        Slow down the animation by a scale.

        Params:
        ----
        - scale: float >= 0, the value by which each frame duration will be divided.

        Raises:
        ----
        - ValueError if the provided scale is negative.
        """
        super().__init__()
        if scale < 0:
            raise ValueError(f"The scale cannot be negative, got {scale}.")
        self.scale = scale

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        new_durations = tuple(d*self.scale for d in durations)
        return surfaces, new_durations, introduction, None, width, height

class SetDurations(Transformation):
    """
    Set the durations of every frame in the art to new values.
    """

    def __init__(self, new_durations: Iterable[int] | int) -> None:
        """
        Set the durations of every frame in the art to new values.

        Params:
        ---
        - new_durations: Iterable[int] | int. If only one int is provided, the duration of all frames is fixed to this value.
        Otherwise, the durations are adjusted one by one. If new_duration has less elements than the number of frame, it is looped.
        If new_durations has more elements than the number of frame, only the first ones are taken.
        """
        super().__init__()
        if not isinstance(new_durations, Iterable):
            new_durations = [new_durations]
        self.new_durations = cycle(new_durations)

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        return surfaces, tuple(next(self.new_durations) for _ in durations), introduction, None, width, height

class SetIntroductionIndex(Transformation):
    """
    Set the introduction to a new index.
    """
    def __init__(self, introduction: int) -> None:
        """
        Set the introduction to a new index.

        Params:
        ---
        - introduction: int, the new index for the introduction.
        """
        super().__init__()
        self.introduction = introduction

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        return surfaces, durations, self.introduction, None, width, height

class SetIntroductionTime(Transformation):
    """
    Set the introduction to a new index by specifying a time.
    """
    def __init__(self, introduction: int) -> None:
        """
        Set the introduction to a new index by specifying a time.

        Params:
        ----
        - introduction: int, a time in ms. Set the introduction to the frame displayed at the given time.
        """
        super().__init__()
        self.introduction = introduction

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):

        sum_dur = 0
        new_intro_idx = 0
        while sum_dur < self.introduction:
            sum_dur += durations[new_intro_idx%len(surfaces)]
            new_intro_idx += 1

        return surfaces, durations, new_intro_idx, None, width, height

def _index_here(index, length, introduction):
    """Return the index on the sequence of frames, looped taking the introduction into account."""
    if index < length: return index
    else: return _index_here(index - length + introduction, length, introduction)

class ExtractSlice(Transformation):
    """This transformation returns a subset of the images and durations of the art."""

    def __init__(self, slice: slice) -> None:
        """
        This transformation returns a subset of the images and durations of the art.
        
        Parmas:
        ---
        - slice: slice, the slice object used to get a slice of the art. If the slice extends further than the number of
        frames, the list of frame is loop, but avoids the frames before the introduction.

        Note:
        ----
        This transformation can be also used by calling art[x] where x is any slice or int, like any list slicing.
        For example:
        art.transform(ExtractSlice(slice(1, 10, 2))) will extract in place, while
        art2 = art[1:10:1] will create a copy with the selected frames.
        """
        super().__init__()
        self.slice = slice

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        # Allow calling for indexing further than the number of frames, in this case, get the frames after having avoided the introduction.
        # Ex: ExtractSlice(slice(10, 19)) on a art with len(art) = 15 and introduction = 7 returns the frames at indices [10, 11, 12, 13, 14, 7, 8, 9]
        indices = range(*self.slice.indices(len(surfaces)*2 - introduction))
        surfaces = tuple(surfaces[_index_here(index, len(surfaces), introduction)] for index in indices)
        durations = tuple(durations[_index_here(index, len(durations), introduction)] for index in indices)
        return surfaces, durations, 0, 0, width, height

class ExtractOne(Transformation):
    """Extract one frame of the animation."""

    def __init__(self, index: int) -> None:
        """
        Extract one frame of the animation.

        Params:
        ---
        - index: int, the index of the frame to extract. If the index is higher than the number of frames, the frames are looped
        avoiding the frames before the introduction.
        """
        super().__init__()
        self.index = index

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        return (surfaces[_index_here(self.index, len(surfaces), introduction)],), (0,), 0, 0, width, height

class First(Transformation):
    """Extract the very first frame of the animation."""

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        return (surfaces[0],), (0,), 0, 0, width, height

class Last(Transformation):
    """Extract the very last frame of the animation."""

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        return (surfaces[-1],), (0,), 0, 0, width, height

class ExtractAtIntroduction(Transformation):
    """Extract the frame of the introduction."""

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        return (surfaces[introduction],), (0,), 0, 0, width, height

class ExtractFromIntroduction(Transformation):
    """Extract the frames after the introduction."""

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        return tuple(surfaces[idx] for idx in range(introduction, len(surfaces))), tuple(0 for _ in range(introduction, len(surfaces))), 0, 0, width, height

class ExtractTime(Transformation):
    """Extract the frame displayed at a given time."""

    def __init__(self, time: int) -> None:
        """
        Extract the frame displayed at a given time.

        Params:
        ---
        time: int, the time at which the targetted frame is displayed. If the time is after the cumulated duration of all frames,
        the frames are looped, avoiding the frames before the introduction.
        """
        super().__init__()
        self.time = time

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        cum_time = 0
        idx = 0
        next_idx = 1

        while self.time >= cum_time + durations[next_idx]:
            cum_time += durations[next_idx]
            idx += 1
            if idx == len(surfaces):
                idx = introduction
            next_idx += 1
            if next_idx == len(surfaces):
                next_idx = introduction
        return (surfaces[idx],), (0,), 0, 0, width, height

class ExtractWindow(Transformation):
    """Extract the frames displayed in a given window of time."""

    def __init__(self, from_time: int, to_time: int):
        """
        Extract the frames displayed in a given window of time.
        
        Params:
        ---
        from_time: int, to_time: int, the limits of the time window, in ms.

        Raises:
        ---
        ValueError of from_time > to_time
        """

        super().__init__()
        if from_time > to_time:
            raise ValueError(f"from_time cannot be greater than to_time, got {from_time} >= {to_time}.")
        self.from_time = from_time
        self.to_time = to_time

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        cum_time = 0
        idx = 0
        next_idx = 1

        # Find the first frame
        while self.from_time >= cum_time + durations[next_idx]:
            cum_time += durations[next_idx]
            idx += 1
            if idx == len(surfaces):
                idx = introduction
            next_idx += 1
            if next_idx == len(surfaces):
                next_idx = introduction

        surfs = [surfaces[idx]]
        durs = [durations[idx]]
        indices = [idx]

        # Find the last frame
        while self.to_time >= cum_time + durations[next_idx]:
            cum_time += durations[next_idx]
            idx += 1
            if idx == len(surfaces):
                idx = introduction
            next_idx += 1
            if next_idx == len(surfaces):
                next_idx = introduction
            surfs.append(surfaces[idx])
            durs.append(durations[idx])
            indices.append(idx)

        if introduction not in indices: introduction = 0
        else: introduction = introduction - min(indices)

        if index not in indices: index = 0
        else: index = index - min(indices)

        return surfs, durs, introduction, index, width, height

class RandomizeIndex(Transformation):
    """Randomize the current index of the animation."""

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        return surfaces, durations, introduction, randint(0, len(surfaces)), width, height

class Shuffle(Transformation):
    """
    Shuffle the order of the frames in the animation.
    
    The index and introduction are set to the new index of the corresponding frame in the shuffled list of frames.
    To avoid having problem with the introduction, consider setting it back to 0 after calling this shuffle.
    """

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        indices = list(range(len(surfaces)))
        shuffle(indices)
        surfaces = tuple(surfaces[idx] for idx in indices)
        durations = tuple(durations[idx] for idx in indices)
        index = indices.index(index)
        introduction = indices.index(introduction)
        return surfaces, durations, introduction, index, width, height



"""The effect module contains transformation consisting on applying effects"""
from typing import Callable
from pygame import Surface, surfarray as sa
import numpy as np
from pygamecv import saturate, desaturate, shift_hue, lighten, darken
from .transformation import Transformation
from ..mask import Mask

class SetAlpha(Transformation):
    """
    The set alpha transformation is used to change the value of the alpha channel.
    """

    def __init__(self, alpha: int = None, mask: Mask = None) -> None:
        """
        If alpha is specified, the SetAlpha transformation replace the alpha value of all the pixel by a new value.
        if mask is specified, the transformation replace the alpha value of all pixel by the value of the mask.
        Pixels that are transparent from the begining will not change.

        Params:
        ---
        - alpha: int, an integer between 0 and 255
        - mask: as gamarts.mask.Mask. One of two must be given.
        """
        super().__init__()

        if alpha is None and mask is None:
            raise ValueError("Both alpha and mask cannot be None.")
        self.alpha = alpha
        self.mask = mask

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        if self.mask is None:
            for surf in surfaces:
                surf.set_alpha(self.alpha)
        else:
            if not self.mask.is_loaded():
                self.mask.load(width, height, **ld_kwargs)
            for surf in surfaces:
                alpha_array = sa.pixels_alpha(surf)
                alpha_array[:] = (1 - self.mask.matrix)*255

        return surfaces, durations, introduction, None, width, height

    def cost(self, width: int, height: int, length: int, **ld_kwargs):
        if self.mask is None:
            return 0
        else:
            return width*height*length

class _MatrixTransformation(Transformation):
    """Matrix transformations are bases for all transformation transforming the matrix with an effect."""

    def __init__(self, mask: Mask | None = None):
        self.mask = mask

    def cost(self, width: int, height:int, length: int, **ld_kwargs):
        if self.mask is None or not self.mask.is_loaded():
            return width*height*length
        # If the mask is not loaded yet, every pixel of something (either the mask or the surfaces) would be impacted.
        # If the mask is already loaded, we get the smallest submask.
        not_null_columns = self.mask.not_null_columns()
        not_null_rows = self.mask.not_null_rows()
        if not_null_columns and not_null_rows:
            return (not_null_columns[-1] - not_null_columns[0])*(not_null_rows[-1]*not_null_rows[0])*length
        return 0

class RBGMap(_MatrixTransformation):
    """
    An RGBMap is a transformation applied directly on the pixel of the surfaces. The alpha value is not taken into account.
    the function must be vectorized. (check numpy.vectorize)
    """

    def __init__(self, function: Callable[[int, int, int], tuple[int, int, int]], mask: Mask = None, mask_threshold: float = 0.99) -> None:
        """
        An RGBMap applies a pixel-by-pixel transformation.

        Params:
        ---
        - function: r,g,b -> (r,g,b). The mapping function.
        - mask: gamarts.mask.Mask, a mask. It is used to know what 
        - mask_threshold: float. The threshold with which the values in the mask is compared. If it is above, then the pixel is
        mapped to a new value. If it is below, the pixel stays unchanged.
        """
        super().__init__(mask)
        self.function = function
        self.mask_threshold = mask_threshold

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        for surf in surfaces:
            rgb_array = sa.pixels3d(surf)
            if self.mask is None:
                rgb_array[:] = np.clip(np.apply_along_axis(lambda t: self.function(*t), 2, rgb_array), 0, 255).astype(np.int8)
            else:
                self.mask.load(width, height, **ld_kwargs)
                rgb_array[self.mask.matrix > self.mask_threshold] = np.clip(np.apply_along_axis(lambda t: self.function(*t), 2, rgb_array[self.mask.matrix > self.mask_threshold]), 0, 255).astype(np.int8)

        return surfaces, durations, introduction, None, width, height

class RGBAMap(_MatrixTransformation):
    """
    An RGBAMap is a transformation applied directly on the pixel of the surfaces. The alpha value is taken into account.
    the function must be vectorized. (check numpy.vectorize)
    """

    def __init__(self, function: Callable[[int, int, int, int], tuple[int, int, int, int]], mask: Mask = None, mask_threshold: float = 0.99) -> None:
        """
        An RGBMap applies a pixel-by-pixel transformation.

        Params:
        ---
        - function: r,g,b,a -> (r,g,b,a). The mapping function.
        - mask: gamarts.mask.Mask, a mask. It is used to know what 
        - mask_threshold: float. The threshold with which the values in the mask is compared. If it is above, then the pixel is
        mapped to a new value. If it is below, the pixel stays unchanged.
        """
        super().__init__(mask)
        self.function = function
        self.mask_threshold = mask_threshold

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):

        for surf in surfaces:

            if self.mask is None:
                rgb_array = sa.pixels3d(surf)
                alpha_array = sa.pixels_alpha(surf)
                r, g, b = rgb_array[:, :, 0], rgb_array[:, :, 1], rgb_array[:, :, 2]
                new_r, new_g, new_b, new_a = self.function(r, g, b, alpha_array)
                rgb_array[:, :, 0] = new_r
                rgb_array[:, :, 1] = new_g
                rgb_array[:, :, 2] = new_b
                alpha_array[:] = new_a
            else:
                self.mask.load(width, height, **ld_kwargs)
                mask = self.mask.matrix > self.mask_threshold
                rgb_array = sa.pixels3d(surf)
                alpha_array = sa.pixels_alpha(surf)[mask]
                r, g, b = rgb_array[:, :, 0][mask], rgb_array[:, :, 1][mask], rgb_array[:, :, 2][mask]
                new_r, new_g, new_b, new_a = self.function(r, g, b, alpha_array)
                rgb_array[:, :, 0][mask] = new_r
                rgb_array[:, :, 1][mask] = new_g
                rgb_array[:, :, 2][mask] = new_b
                alpha_array[:] = new_a

        return surfaces, durations, introduction, None, width, height

class Saturate(_MatrixTransformation):
    """Saturate the art by a given factor."""

    def __init__(self, factor: float, mask: Mask = None) -> None:
        """
        Saturate the frames of the art by a given factor.

        Params:
        ---
        - factor: float, 0 <= factor <= 1. The factor by which the frames will be saturated.
        - mask: mask.Mask. If given, the saturation factor applied is mask.matrix*factor. The pixels will be saturated one by one.
        """
        super().__init__(mask)
        self.factor = factor

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        if self.mask is not None:
            self.mask.load(width, height, **ld_kwargs)
            factor = self.mask.matrix*self.factor
        else:
            factor = self.factor
        for surf in surfaces:
            saturate(surf, factor)

        return surfaces, durations, introduction, None, width, height
    
class Desaturate(_MatrixTransformation):
    """Desaturate the art by a given factor."""

    def __init__(self, factor: float, mask: Mask = None) -> None:
        """
        Desaturate the frames of the art by a given factor.

        Params:
        ---
        - factor: float, 0 <= factor <= 1. The factor by which the frames will be desaturated.
        - mask: mask.Mask. If given, the desaturation factor applied is mask.matrix*factor. The pixels will be desaturated one by one.
        """
        super().__init__(mask)
        self.factor = factor

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        if self.mask is not None:
            self.mask.load(width, height, **ld_kwargs)
            factor = self.mask.matrix*self.factor
        else:
            factor = self.factor
        for surf in surfaces:
            desaturate(surf, factor)

        return surfaces, durations, introduction, None, width, height
    
class Darken(_MatrixTransformation):
    """Darken the art by a given factor."""

    def __init__(self, factor: float, mask: Mask = None) -> None:
        """
        Darken the frames of the art by a given factor.

        Params:
        ---
        - factor: float, 0 <= factor <= 1. The factor by which the frames will be darkened.
        - mask: mask.Mask. If given, the darkening factor applied is mask.matrix*factor. The pixels will be darkened one by one.
        """
        super().__init__(mask)
        self.factor = factor

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        if self.mask is not None:
            self.mask.load(width, height, **ld_kwargs)
            factor = self.mask.matrix*self.factor
        else:
            factor = self.factor
        for surf in surfaces:
            darken(surf, factor)

        return surfaces, durations, introduction, None, width, height

class Lighten(_MatrixTransformation):
    """Lighten the art by a given factor."""

    def __init__(self, factor: float, mask: Mask = None) -> None:
        """
        Lighten the frames of the art by a given factor.

        Params:
        ---
        - factor: float, 0 <= factor <= 1. The factor by which the frames will be lightened.
        - mask: mask.Mask. If given, the lightening factor applied is mask.matrix*factor. The pixels will be lightened one by one.
        """
        super().__init__(mask)
        self.factor = factor

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        if self.mask is not None:
            self.mask.load(width, height, **ld_kwargs)
            factor = self.mask.matrix*self.factor
        else:
            factor = self.factor
        for surf in surfaces:
            lighten(surf, factor)

        return surfaces, durations, introduction, None, width, height

class ShiftHue(_MatrixTransformation):
    """Shift the hue of all surface of the art by a given value."""

    def __init__(self, value: int, mask: Mask = None) -> None:
        """
        shirt of the hue of the frames of the art by a given value.

        Params:
        ---
        - value: int. The value by which the hue will be shifted.
        - mask: mask.Mask. If given, the shifting value applied is mask.matrix*value. The pixels will be shifted one by one.
        """
        super().__init__(mask)
        self.value = value

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        if self.mask is not None:
            self.mask.load(width, height, **ld_kwargs)
            value = self.mask.matrix*self.value
        else:
            value = self.value
        for surf in surfaces:
            shift_hue(surf, value/2)

        return surfaces, durations, introduction, None, width, height

class Invert(_MatrixTransformation):
    """Invert the color of the art."""

    def __init__(self, mask: Mask = None, mask_threshold: float = 0.99):
        """
        Invert the color of the art.
        
        Params:
        ----
        - mask: mask.Mask. If specified, the mask is used to know what pixels are to be inverted.
        - mask_threshold: float. the threshold used to convert the mask into a boolean mask. Pixels whose value in the mask
        is above the threshold will be inverted. The other won't.
        """
        super().__init__(mask)
        self.mask_threshold = mask_threshold

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        for surf in surfaces:
            rgb_array = sa.pixels3d(surf)
            if self.mask is None:
                rgb_array[:] = 255 - rgb_array
            else:
                self.mask.load(width, height, **ld_kwargs)
                rgb_array[self.mask.matrix > self.mask_threshold] = 255 - rgb_array[self.mask.matrix > self.mask_threshold]
        return surfaces, durations, introduction, None, width, height

class AdjustContrast(_MatrixTransformation):
    """Change the contrast of an art. The constrast is a value between -255 and +255."""

    def __init__(self, contrast: int, mask: Mask = None, mask_threshold: float = 0.99) -> None:
        """
        Change the contrast of an art. The constrast is a value between -255 and +255

        Params:
        ----
        - mask: mask.Mask. If specified, the mask is used to know what pixels are to be adjusted.
        - mask_threshold: float. the threshold used to convert the mask into a boolean mask. Pixels whose value in the mask
        is above the threshold will be adjusted. The other won't.
        """
        
        super().__init__(mask)
        self.factor = (259 * (contrast + 255)) / (255 * (259 - contrast))
        self.mask_threshold = mask_threshold

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        for surf in surfaces:
            rgb_array = sa.pixels3d(surf)
            if self.mask is None:
                rgb_array[:] = np.clip((self.factor * (rgb_array - 128) + 128).astype(np.int8), 0, 255)
            else:
                self.mask.load(width, height, **ld_kwargs)
                rgb_array[self.mask.matrix > self.mask_threshold] = np.clip(
                    (self.factor * (rgb_array[self.mask.matrix > self.mask_threshold] - 128) + 128).astype(np.int8), 0, 255)

        return surfaces, durations, introduction, None, width, height

class AddBrightness(_MatrixTransformation):
    """Change the brightness of an art. The brightness is a value between -255 and +255."""

    def __init__(self, brightness: int, mask: Mask = None, mask_threshold: float = 0.99) -> None:
        """
        Change the brightness of an art. The brightness is a value between -255 and +255.
        Adding brightness to the art will just add a fixed value to all the pixels.
        It is equivalent to use a RGBAMap with the function lambda r,g,b : (r+B, g+B, b+B) where B is the brightness.

        Params:
        ----
        - mask: mask.Mask. If specified, the mask is used to know what pixels are to be adjusted.
        - mask_threshold: float. the threshold used to convert the mask into a boolean mask. Pixels whose value in the mask
        is above the threshold will be adjusted. The other won't.
        """
        super().__init__(mask)
        self.brightness = np.int8(brightness)
        self.mask_threshold = mask_threshold

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        for surf in surfaces:
            rgb_array = sa.pixels3d(surf)
            if self.mask is None:
                rgb_array[:] = np.clip(rgb_array + self.brightness, 0, 255).astype(np.int8)
            else:
                if not self.mask.is_loaded():
                    self.mask.load(width, height, **ld_kwargs)
                rgb_array[self.mask.matrix > self.mask_threshold] = np.clip(rgb_array[self.mask.matrix > self.mask_threshold] + self.brightness, 0, 255)
        return surfaces, durations, introduction, None, width, height

class Gamma(_MatrixTransformation):
    """
    The gamma transformation is used to modify the brightness of the image.
    For 0 < gamma < 1, the dark pixels will be brighter and the bright pixels will not change
    For gamma > 1, the light pixels will be darker and the dark pixel will not change
    """

    def __init__(self, gamma: float, mask: Mask = None, mask_threshold: float = 0.99) -> None:
        """
        Apply a gamma transformation.

        Params:
        ----
        - gamma: the parameter of the transformation.
        - mask: mask.Mask. If specified, the mask is used to know what pixels are to be adjusted.
        - mask_threshold: float. the threshold used to convert the mask into a boolean mask. Pixels whose value in the mask
        is above the threshold will be adjusted. The other won't.
        """
        super().__init__(mask)
        self.gamma = gamma
        self.mask_threshold = mask_threshold  

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        for surf in surfaces:
            rgb_array = sa.pixels3d(surf)
            if self.mask is None:
                rgb_array[:] = np.clip(((rgb_array/255)**self.gamma * 255).astype(np.int8), 0, 255)
            else:
                if not self.mask.is_loaded():
                    self.mask.load(width, height, **ld_kwargs)
                rgb_array[self.mask.matrix > self.mask_threshold] = np.clip(
                    ((rgb_array[self.mask.matrix > self.mask_threshold]/255)**self.gamma * 255).astype(np.int8), 0, 255)

        return surfaces, durations, introduction, None, width, height



"""The transform module contiains all the transformations that can be applied to an Art."""
from .combination import Blit, Average, Concatenate
from .transformation import (
    Transformation, Pipeline,
    SetIntroductionIndex, SetIntroductionTime, SlowDown, SpeedUp, SetDurations,
    Resize, Rotate, Crop, VerticalChop, HorizontalChop, Last, ExtractSlice, ExtractOne, First, Flip, Transpose,
    Zoom, Pad, ExtractTime, ExtractWindow
)
from .drawing import DrawArc, DrawCircle, DrawEllipse, DrawLine, DrawLines, DrawPie, DrawPolygon, DrawRectangle, DrawRoundedRectantle
from .effect import Saturate, Darken, Lighten, Desaturate, SetAlpha, ShiftHue, Gamma, AdjustContrast, RBGMap, RGBAMap, Invert, AddBrightness
from .convert import GrayScale, ConvertRGB, ConvertRGBA



"""The transformation submodule contains all masks being transformations of an image, array or of another mask."""
from abc import ABC, abstractmethod
from typing import Callable, Sequence
from  pygame import surfarray as sa, transform as tf, image as im
import numpy as np
from .mask import Mask


class FromArtAlpha(Mask):
    """A mask from the alpha layer of an art."""

    def __init__(self, art, index: int= 0) -> None:
        """
        A mask from the alpha layer of an art.
        
        Params:
        ---
        - art: Art, the art whose alpha layer is used. A rescaled copy is used if the width and height of the art isn't matching the requested width and height.
        - index: int. The index of the frame to be used in the art.

        Notes:
        ---
        - If the art isn't loaded when the loading of this mask happens, the art will be loaded, then unloaded.
        """
        super().__init__()
        self.art = art
        self.index = index

    def _load(self, width: int, height: int, **ld_kwargs):
        need_to_unload = False
        if not self.art.is_loaded():
            need_to_unload = True
            self.art.load(**ld_kwargs)

        self.matrix = 1 - sa.array_alpha(tf.scale(self.art.surfaces[self.index], (width, height)).convert_alpha()).astype(np.int8)/255

        if need_to_unload:
            self.art.unload()

class FromArtColor(Mask):
    """
    A mask from a mapping of the color layers.
    
    Every pixel of the art is mapped to a value between 0 and 1 with the provided function.
    Selects only one image of the art based on the index.
    """

    def __init__(self, art, function: Callable[[int, int, int], float], index: int = 0) -> None:
        """
        A mask from the color layers of an art.
        
        Params:
        ---
        - art: Art, the art whose color layers is used. A rescaled copy is used if the width and height of the art isn't matching the requested width and height.
        - function: Callable[[int, int, int], float]: A function mapping an rgb tuple to a float between 0 and 1.
        - index: int. The index of the frame to be used in the art.

        Notes:
        ---
        - If the art isn't loaded when the loading of this mask happens, the art will be loaded, then unloaded.
        """
        super().__init__()
        self.art = art
        self.index = index
        self.map = function

    def _load(self, width: int, height: int, **ld_kwargs):
        need_to_unload = False
        if not self.art.is_loaded():
            need_to_unload = True
            self.art.load(**ld_kwargs)

        self.matrix = np.apply_along_axis(lambda t: self.map(*t), 2, sa.array3d(tf.scale(self.art.surfaces[self.index], (width, height))).astype(np.int64))

        if need_to_unload:
            self.art.unload()

class FromImageColor(Mask):
    """
    A mask from an image.
    
    Every pixel of the art is mapped to a value between 0 and 1 with the provided function.
    """

    def __init__(self, path: str, function: Callable[[int, int, int], float]) -> None:
        """
        A mask from an image.

        Params:
        ---
        - path: the path to the image.
        - function: Callable[[int, int, int], float]: A function mapping an rgb tuple to a float between 0 and 1.
        """
        self.path = path
        super().__init__()
        self.map = function

    def _load(self, width: int, height: int, **ld_kwargs):
        rgb_array = sa.array3d(tf.scale(im.load(self.path), (width, height)))
        self.matrix = np.apply_along_axis(lambda t: self.map(*t), 2, rgb_array.astype(np.int64))

class _MaskCombination(Mask, ABC):
    """MaskCombinations is an abstract class for all mask combinations: sum, products and average"""

    def __init__(self, *masks: Mask):
        """
        Combine the masks.

        Params:
        ---
        - *masks: Mask, the masks to be combined.
        """

        super().__init__()
        self.masks = masks

    @abstractmethod
    def _combine(self, *matrices: np.ndarray) -> np.ndarray:
        raise NotImplementedError()

    def _load(self, width:int, height: int, **ld_kwargs):
        for mask in self.masks:
            if not mask.is_loaded():
                mask.load(width, height, **ld_kwargs)

        self.matrix = self._combine(*(mask.matrix for mask in self.masks))

class SumOfMasks(_MaskCombination):
    """
    A sum of masks is a mask based on the sum of the matrixes of the masks, clamped between 0 and 1.
    For binary masks, it acts like union.
    """

    def _combine(self, *matrices):
        return np.clip(sum(matrices), 0, 1)

class DifferenceOfMasks(_MaskCombination):
    """
    A difference of masks is a masked based on the difference between the matrixes of two masks, clamped between 0 and 1.
    """
    # pylint: disable=arguments-differ
    def _combine(self, matrix1, matrix2):
        return matrix1 - matrix2

class ProductOfMasks(_MaskCombination):
    """
    A product of masks is a mask based on the product of the matrixes of the masks.
    For binary masks, it acts like intersections.
    """

    def _combine(self, *matrices):
        prod = 1
        for mat in matrices:
            prod*=mat
        return prod

class AverageOfMasks(_MaskCombination):
    """
    An average of masks is a mask based on the average of the matrixes of the masks.
    """

    def __init__(self, *masks: Mask, weights: Sequence = None):
        """
        An average of masks is a mask based on the average of the matrixes of the masks.

        Params:
        ---
        - *masks: Mask, the masks to be averaged.
        - weights: Sequence, a sequence of number to be used to weight the average.

        Raises:
        ---
        ValueError if the weights are provded and do not have a length equa to the number of masks.
        """
        if weights is None:
            weights = [1]*len(masks)
        if len(weights) != len(masks):
            raise ValueError("The number of weights should match the number of masks")
        super().__init__(*masks)
        self.weights = weights

    def _combine(self, *matrices):
        avg = 0
        for matrix, weight in zip(matrices, self.weights):
            avg += matrix*weight

        avg /= sum(self.weights)
        return avg

class BlitMaskOnMask(_MaskCombination):
    """
    A blit mask on mask is a mask where the values of the background below (or above) a given threshold are replaced
    by the values on the foreground that are below a given threshold.
    """

    def __init__(self, background: Mask, foreground: Mask, bg_threshold: float = 0, fg_threshold: float = 0.5, reverse: bool = False):
        """
        Replace the values from a mask by others.
        
        Params:
        ---
        - background: mask, the mask having some values to be replaced.
        - foreground: mask, the mask whose values will replace the background values.
        - bg_threshold: Only the value of the background above (or below if reverse is True) that threshold are replaced.
        - fg_threshold: Only the value of the foreground below that threshold are used.
        - reverse: flag for comparison with bg_threshold.
        """
        super().__init__(background, foreground)
        self.bg_threshold = bg_threshold
        self.fg_threshold = fg_threshold
        self.reverse = reverse
    #pylint: disable=arguments-differ
    def _combine(self, background_matrix, foreground_matrix) -> np.ndarray:
        matrix = background_matrix
        if self.reverse:
            positions_to_keep = (background_matrix < self.bg_threshold) & (foreground_matrix < self.fg_threshold)
        else:
            positions_to_keep = (background_matrix > self.bg_threshold) & (foreground_matrix < self.fg_threshold)
        matrix[positions_to_keep] = foreground_matrix[positions_to_keep]
        return matrix

class InvertedMask(Mask):
    """
    An inverted mask is a mask whose value are the opposite of the parent mask.
    """

    def __init__(self, mask: Mask):
        """
        An inverted mask is a mask whose value are the opposite of the parent mask.

        Params:
        ---
        - mask: Mask, the mask to be inverted.
        """
        super().__init__()
        self._mask = mask

    def _load(self, width:int, height: int, **ld_kwargs):
        if not self._mask.is_loaded():
            self._mask.load(width, height, **ld_kwargs)
        self.matrix = 1 - self._mask.matrix

class TransformedMask(Mask):
    """
    A Transformed mask is a mask whose matrix is the transformation of the matrix of another mask.
    The transformation must be a numpy vectorized function or a function matrix -> matrix.
    """

    def __init__(self, mask: Mask, transformation: Callable[[float], float] | Callable[[np.ndarray], np.ndarray]):
        """
        Create a mask from the transformation of another mask.
        
        Params:
        ---
        - mask: Mask, the mask to be transformed
        - transformation: Callable[[float], float] | Callable[[numpy.ndarray], numpy.ndarray]: The transformation applied to the mask.
        The transformation create an array from another. Both arrays must have the same shape.
        """
        super().__init__()
        self._mask = mask
        self.transformation = transformation

    def _load(self, width:int, height: int, **ld_kwargs):
        if not self._mask.is_loaded():
            self._mask.load(width, height, **ld_kwargs)

        self.matrix = np.clip(self.transformation(self._mask.matrix), 0, 1)
        if self.matrix.shape != self._mask.matrix.shape:
            raise ValueError(f"Shape of the mask changed from {self._mask.matrix.shape} to {self.matrix.shape}")

class BinaryMask(Mask):
    """
    A binary mask is a mask where every values are 0 or 1. It is based on another mask.
    The matrix of this mask is that every component is 1 if the value on the parent mask
    is above a thresold and 0 otherwise. (this is reversed if reverse is set to True).
    """

    def __init__(self, mask: Mask, threshold: float, reverse: bool = False):
        """
        Create a binary mask.
        
        Params:
        ---
        - mask: the mask to be binarized
        - threshold: the threshold used to compare the values.
        - reverse: flag used to know if the 1 of the binary mask are above the threshold (not reversed) or below the threshold (reversed)
        """

        super().__init__()
        self.threshold = threshold
        self._mask = mask
        self.reverse = reverse

    def _load(self, width:int, height: int, **ld_kwargs):
        if not self._mask.is_loaded():
            self._mask.load(width, height, **ld_kwargs)

        if self.reverse:
            positions_to_keep = self._mask.matrix < self.threshold
        else:
            positions_to_keep = self._mask.matrix > self.threshold

        self.matrix = np.zeros_like(self._mask.matrix)
        self.matrix[positions_to_keep] = 1



"""The file module contains classes to open images, gifs and folders as arts."""
from typing import Iterable
import os
from PIL import Image
from pygame.image import load, fromstring
from .art import Art
from .._common import LoadingError
from ..transform import Transformation

class ImageFile(Art):
    """
    The ImageFile class is an Art loaded from an image.
    Accepted format are: jpg, jpeg, png, gif (only first frame), svg, webp, lmb, pcx, pnm, tga (uncompressed), xpm

    Example:
    ---
    - ``ImageFile("my_image.png")`` is an Art displaying the image stored at "my_image.png"
    - ``ImageFile("characters/char1.jpeg", False)`` will load an the image stored at "characyers/char1.jpeg" and convert it in the RGB format
    """

    def __init__(self, file: str, transparency: bool = True, transformation: Transformation = None) -> None:
        """
        The ImageFile class is an Art loaded from an image.
        Accepted format are: jpg, jpeg, png, gif (only first frame), svg, webp, lmb, pcx, pnm, tga (uncompressed), xpm
        
        Params:
        ----
        - file: str, the path to the file.
        - transparency: bool = True, whether the surface has some transparency. ImageFile with transparency are in the RGBA format, while ImageFiles
        without are in the RGB format (without alpha channel).
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        """
        super().__init__(transformation)
        self.full_path = file
        im = Image.open(self.full_path)
        self._width, self._height = im.size
        self._transparency = transparency
        im.close()
        self._find_initial_dimension()

    def _load(self, **ld_kwargs):
        if self._transparency: self._surfaces = (load(self.full_path).convert_alpha(),)
        else: self._surfaces = (load(self.full_path).convert(),)
        self._durations = (0,)

class ImageFolder(Art):
    """
    The ImageFolder class is an Art loaded from multiple images in a folder.
    All image must have one of these formats: jpg, jpeg, png, gif (only first frame), svg, webp, lmb, pcx, pnm, tga (uncompressed), xpm
    The animation is reconstructed by taking images in the alphabetical order from the file. All images must have the same sizes
    
    Example:
    -----
    - ``ImageFolder("my_images/", [100, 200, 100])``
    is an Art displaying the images stored in the folder "assets/images/my_images/".
    The folder contains 3 images and the animation will be 400 ms, 100 ms for the first image, 200 ms for the second, and 100 for the last
    - ``ImageFolder("characters/char1/running/", 70)`` is an Art displaying the images stored in the folder "assets/images/characters/char1/running/".
    Every images in the folder will be display 70 ms.
    - ``ImageFolder("my_images/", 70, 5)`` is an Art displaying the images stored in the folder "assets/images/my_images/".
    The folder must contains at least 5 images.
    When all the images have been displayed, it does not loop on the very first but on the 6th.
    Frame will be displayed in the following order, if there are 9 frames:
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 5, 6, 7, 8, 5, 6, 7, 8, 5, 6, ...].
    """

    def __init__(
        self,
        folder: str,
        durations: Iterable[int] | int,
        introduction: int = 0,
        transformation: Transformation = None,
    ) -> None:
        """
        The ImageFolder class is an Art loaded from multiple images in a folder.
        All image must have one of these formats: jpg, jpeg, png, gif (only first frame), svg, webp, lmb, pcx, pnm, tga (uncompressed), xpm
        The animation is reconstructed by taking images in the alphabetical order from the file. All images must have the same sizes.

        Params:
        ---
        - folder: str, the path to the folder.
        - durations: Iterable[int] | int, the duration(s) of the frames. If an iterable is provided, it must have the same number of elements than
        the number of images in the folder.
        - introduction: the introduction of the art. If specified, the art will not display the first frame after the last but this one instead. See examples.
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        
        Raises:
        ---
        - LoadingError if the specified introduction is larger than the number of images in the folder.
        - LoadingError if the iterable of durations does not have the same length as the number of images in the folder.
        """
        super().__init__(transformation)
        self.full_path = folder
        self.durs = durations
        self._introduction = introduction

        self._paths = [
            os.path.join(self.full_path, f)
            for f in os.listdir(self.full_path)
            if os.path.isfile(os.path.join(self.full_path, f))
        ]
        im = Image.open(self._paths[0])
        self._width, self._height = im.size
        im.close()
        self._find_initial_dimension()

    def _load(self, **ld_kwargs):
        self._surfaces = tuple(load(path) for path in self._paths)
        if self._introduction > len(self._paths):
            raise LoadingError(
                f"The introduction specified for this ImageFolder is too high, got {self._introduction} while there is only {len(self.surfaces)} images."
            )
        if isinstance(self.durs, int):
            self._durations = tuple(self.durs for _ in self._paths)
        else:
            if len(self.durs) != len(self._paths):
                raise LoadingError(
                    f"The length of the durations list ({len(self.durs)}) does not match the len of the number of images ({len(self.surfaces)})"
                )
            self._durations = tuple(self.durs)
        self._verify_sizes()

class GIFFile(Art):
    """
    The GIFFile is an Art that displays a gif.
    
    Example:
    -----
    - GIFFile("my_animation.gif") is an Art displaying the gif stored at "assets/images/my_animation.gif".
    - GIFFile("my_animation.gif", 10) is an Art displaying the gif stored at "assets/images/my_animation.gif".
    it must have at least 10 images.
    When all the images have been displayed, do not loop on the very first but on the 10th.
    """

    def __init__(self, file: str, introduction: int = 0, transformation: Transformation = None) -> None:
        """
        The GIFFile is an Art that displays a gif.

        Params:
        ----
        - file: str, the path to the .gif file.
        - introduction: the introduction of the art. If specified, the art will not display the first frame after the last but this one instead. See examples.
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        
        Raises:
        ---
        - LoadingError if the specified introduction is larger than the number of images in the folder.
        """
        super().__init__(transformation)
        self.full_path = file
        self._introduction = introduction
        im = Image.open(self.full_path)
        self._width, self._height = im.size
        im.close()
        self._find_initial_dimension()

    def _load(self, **ld_kwargs):
        gif = Image.open(self.full_path)
        gif.seek(0)
        images = [fromstring(gif.convert('RGBA').tobytes(), gif.size, 'RGBA')]
        image_durations = [gif.info['duration']]
        while True:
            try:
                gif.seek(gif.tell()+1)
                images.append(fromstring(gif.convert('RGBA').tobytes(), gif.size, 'RGBA'))
                image_durations.append(gif.info['duration'])
            except EOFError:
                break
        self._surfaces = tuple(images)
        self._durations = tuple(image_durations)

        if self._introduction > len(self.surfaces):
            raise LoadingError(
                f"The introduction specified for this GIFFile is too high, got {self._introduction} while there is only {len(self.surfaces)} images."
            )
        gif.close()



"""This mask submodule contains the bases for masks and geometrical masks."""
from typing import Sequence
from abc import ABC, abstractmethod
from ZOCallable import ZOZOCallable, verify_ZOZOCallable
from ZOCallable.functions import linear
import numpy as np
from pygame import Surface, surfarray as sa, SRCALPHA, draw, Rect
from .._common import LoadingError

class Mask(ABC):
    """Mask is an abstract class for all masks."""

    def __init__(self) -> None:
        super().__init__()
        self._loaded = False
        self.matrix: np.ndarray = None

    @abstractmethod
    def _load(self, width: int, height: int, **ld_kwargs):
        raise NotImplementedError()

    def load(self, width: int, height: int, **ld_kwargs):
        """
        Load the mask. If the mask is already loaded, do nothing.
        
        Params:
        ----
        - width, height: the dimension of the mask.
        - **ld_kwargs: the loading kwargs.
        """
        if not self._loaded:
            self._load(width, height,**ld_kwargs)
            self._loaded = True

    def unload(self):
        """Unload the mask."""
        self.matrix = None
        self._loaded = False

    def is_loaded(self):
        """Return True if the mask is loaded, False otherwise."""
        return self._loaded

    def not_null_columns(self):
        """
        Return the list of indices of the columns that have at least one value different from 0.
        
        Returns:
        ----
        - list[int], the list of indices of the columns having at least one non-zero value.

        Raises:
        ----
        LoadingError if the mask isn't loaded yet.
        """
        if self.is_loaded():
            return np.where(self.matrix.any(axis=0))[0]
        raise LoadingError("Unloaded masks do not have any matrix and so not null columns.")

    def not_null_rows(self):
        """Return the list of indices of the rows that have at least one value different from 0."""
        if self.is_loaded():
            return np.where(self.matrix.any(axis=1))[0]
        raise LoadingError("Unloaded masks do not have any matrix and so not null rows.")

    def is_empty(self):
        """Return True if all the pixels in the mask are set to 0."""
        if self.is_loaded():
            return np.sum(self.matrix) == 0
        raise LoadingError("Unloaded masks cannot be empty.")

    def is_full(self):
        """Return True if all the pixels in the mask are set to 1."""
        if self.is_loaded():
            return np.sum(self.matrix) == self.matrix.size
        raise LoadingError("Unloaded masks cannot be full.")

class MatrixMask(Mask):
    """
    A matrix mask is a mask based on a matrix, this matrix can be changed.
    When loaded or updated, the matrix of the mask is padded or cropped to fit the requested width and height.
    """

    def __init__(self, matrix: np.ndarray) -> None:
        """
        A matrix mask is a mask based on a matrix, this matrix can be changed.
        When loaded or updated, the matrix of the mask is padded or cropped to fit the requested width and height, and clipped in [0, 1]

        Params:
        - matrix: np.ndarray: the matrix on which the mask is based
        """

        super().__init__()
        self._matrix = np.clip(matrix, 0, 1)

    def _load(self, width: int, height: int, **ld_kwargs):
        """Pad and crop the matrix to match the size."""
        self.matrix = np.pad(
            self._matrix, [
                (0, max(0, width - self._matrix.shape[0])),
                (0, max(0, height - self._matrix.shape[1]))
            ],
            'edge'
        )[: width, :height]

    def update_matrix(self, new_matrix: np.ndarray):
        """Update the current matrix with a user-defined matrix.
            The matrix is padded or cropeed to it the requested width and height.
        """
        self._matrix = new_matrix
        self._load(*self.matrix.shape)

class Circle(Mask):
    """A Circle is a mask with two values: 0 inside the circle and 1 outside."""

    def __init__(self, radius: float | int, center: tuple[float, float] | tuple[int, int] = (0.5, 0.5)):
        """A Circle is a mask with two values: 0 inside the circle and 1 outside.
        
        Params:
        ---
        - radius: int | float, the radius of the circle. If a float between 0 and 1 is provided, it refers to a fraction of min(width, height).
        For example, if the mask is loaded with a size of (100, 150) and radius = 0.3, the mask contains a circle of radius 30.
        If an int is provided, use it as the radius.
        - center: tuple[int, int] | tuple[float, float]. If two floats between 0 and 1 are provided, they refer to a fraction of width and height.
        For example, if the mask is loaded with a size of (100, 150) and a center of 0.3, 0.5, the mask contains a circle centered in 30, 75.
        If ints are provided, they are directly used. The two values are independant, an int can be provided for the x component and a float for
        the second component (and vice versa).
        """
        super().__init__()
        self.radius = radius
        self.center = center

    def _load(self, width: int, height: int, **ld_kwargs):
        grid_x, grid_y = np.ogrid[:width, :height]
        center = (
            self.center[0]*width if 0 <= self.center[0] <= 1 else self.center[0],
            self.center[1]*height if 0 <= self.center[1] <= 1 else self.center[1]
        )
        distances = np.sqrt((grid_x - center[0] - 0.5) ** 2 + (grid_y - center[1] - 0.5) ** 2)
        radius = self.radius*min(height, width) if  0 <= self.radius <= 1 else self.radius
        self.matrix = (distances > radius).astype(int)

class Ellipse(Mask):
    """An Ellipse is a mask with two values: 0 in the ellipse and 1 outside."""

    def __init__(self, radius_x: int | float, radius_y: int | float, center: tuple[int, int] | tuple[float, float] = (0.5, 0.5)):
        """An Ellipse is a mask with two values: 0 in the ellipse and 1 outside.
        
        Params:
        ---
        - radius_x: int | float, the horizontal radius of the ellipse. If a float between 0 and 1 is provided, it refers to a fraction of the width.
        - radius_y: int | float, the vertical radius of the ellipse. If a float between 0 and 1 is provided, it refers to a fraction of the height.
        For example, if the mask is loaded with a size of (100, 150) and radius_x = 0.3, radius_y = 50, the mask contains an ellipse of radii 30, 50.
        If an int is provided, use it as the radius.
        - center: tuple[int, int] | tuple[float, float]. If two floats between 0 and 1 are provided, they refer to a fraction of width and height.
        For example, if the mask is loaded with a size of (100, 150) and a center of 0.3, 0.5, the mask contains a ellispe centered in 30, 75.
        If ints are provided, they are directly used. The two values are independant, an int can be provided for the x component and a float for
        the second component (and vice versa).
        """
        super().__init__()
        self.radius_x = radius_x
        self.radius_y = radius_y
        self.center = center

    def _load(self, width: int, height: int, **ld_kwargs):
        grid_y, grid_x = np.ogrid[:height, :width]
        center = (
            self.center[0]*width if 0 <= self.center[0] <= 1 else self.center[0],
            self.center[1]*height if 0 <= self.center[1] <= 1 else self.center[1]
        )
        radius_x = self.radius_x*width if  0 <= self.radius_x <= 1 else self.radius_x
        radius_y = self.radius_y*height if  0 <= self.radius_y <= 1 else self.radius_y
        distances = np.sqrt((grid_x - center[0] - 0.5) ** 2 / radius_x**2 + (grid_y - center[1] - 0.5) ** 2 / radius_y**2)

        self.matrix = (distances > 1).astype(int)

class Rectangle(Mask):
    """A Rectangle is a mask with two values: 0 inside the rectangle and 1 outside."""

    def __init__(self, left: int | float, top: int | float, right: int | float, bottom: int | float):
        """
        A Rectangle is a mask with two values: 0 inside the rectangle and 1 outside.
        
        Params:
        ----
        - left: int | float, the coordinate of the left of the rectangle, included.
        - top: int | float, the coordinate of the top of the rectangle, included.
        - right: int | float, the coordinate of the right of the rectangle, included.
        - bottom: int | float, the coordinate of the bottom of the rectangle, included.

        For the four arguments, if a float is provided, it is used as a fraction of the width or height

        Example:
        ----
        >>> r = Rectangle(2, 1, 4, 5)
        >>> r.load(6, 4)
        >>> print(r.matrix)
        [[1 1 1 1 1 1]
         [1 1 0 0 0 1]
         [1 1 0 0 0 1]
         [1 1 1 1 1 1]]

        >>> r = Rectangle(0.5, 0.5, 1, 1)
        >>> r.load(6, 4)
        >>> print(r.matrix)
        [[0 0 0 0 0 0]
         [0 0 0 0 0 0]
         [0 0 0 1 1 1]
         [0 0 0 1 1 1]]
        """

        super().__init__()
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom

    def _load(self, width:int, height:int, **ld_kwargs):
        grid_y, grid_x = np.ogrid[:height, :width]
        left = self.left*width if 0 <= self.left <= 1 and isinstance(self.left, float) else self.left
        right = self.right*width if 0 <= self.right <= 1 and isinstance(self.right, float) else self.right
        top = self.top*height if 0 <= self.top <= 1 and isinstance(self.top, float) else self.top
        bottom = self.bottom*height if 0 <= self.bottom <= 1 and isinstance(self.bottom, float) else self.bottom
        self.matrix = 1 - ((left <= grid_x) & (grid_x <= right) & (top <= grid_y) & (grid_y <= bottom)).astype(int)

class Polygon(Mask):
    """
    A Polygon is a mask with two values: 0 inside the polygon and 1 outside the polygon.
    The Polygon is defined from a list of points. If points are outside of [0, width] x [0, height],
    the polygon is cropped.
    """

    def __init__(self, points: Sequence[tuple[int, int]]) -> None:
        """
        A Polygon is a mask with two values: 0 inside the polygon and 1 outside the polygon.
        The Polygon is defined from a list of points. If points are outside of [0, width] x [0, height],
        the polygon is cropped.

        Params:
        ----
        points: the sequence of points used to draw th polygon.
        """
        super().__init__()
        self.points = points

    def _load(self, width: int, height: int, **ld_kwargs):
        surf = Surface((width, height), SRCALPHA)
        draw.polygon(surf, (0, 0, 0, 255), self.points)
        self.matrix = 1 - sa.array_alpha(surf)/255

class RoundedRectangle(Mask):
    """A RoundedRectangle mask is a mask with two values: 0 inside of the rectangle with rounded vertexes, and 1 outside."""

    def __init__(self, left: int | float, top: int | float, right: int | float, bottom: int | float, radius: int):

        """
        A Rectangle is a mask with two values: 0 inside the rectangle and 1 outside. The corners are rounded.
        
        Params:
        ----
        - left: int | float, the coordinate of the left of the rectangle, included.
        - top: int | float, the coordinate of the top of the rectangle, included.
        - right: int | float, the coordinate of the right of the rectangle, included.
        - bottom: int | float, the coordinate of the bottom of the rectangle, included.
        - radius: int, the radius of the corner.

        For the four arguments, if a float is provided, it is used as a fraction of the width or height.
        """

        super().__init__()
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom
        self.radius = radius

    def _load(self, width: int, height: int, **ld_kwargs):
        surf = Surface((width, height), SRCALPHA)
        left = self.left*width if 0 <= self.left <= 1 and isinstance(self.left, float) else self.left
        right = self.right*width if 0 <= self.right <= 1 and isinstance(self.right, float) else self.right
        top = self.top*height if 0 <= self.top <= 1 and isinstance(self.top, float) else self.top
        bottom = self.bottom*height if 0 <= self.bottom <= 1 and isinstance(self.bottom, float) else self.bottom
        draw.rect(surf, (0, 0, 0, 255), Rect(left, top, right - left + 1, bottom - top + 1), 0, self.radius)
        self.matrix = 1 - sa.array_alpha(surf)/255

class GradientCircle(Mask):
    """
    A GradientCircle mask is a mask where the values ranges from 0 to 1. All pixels in the inner circle are set to 0,
    all pixels out of the outer cirlce are set to 1, and pixels in between have an intermediate value.

    The intermediate value is defined by the transition function. This function must be vectorized.
    """

    def __init__(
            self,
            inner_radius: int | float,
            outer_radius: int | float = 0.5,
            transition: ZOZOCallable = linear,
            center: tuple[float, float] | tuple[int, int] = (0.5, 0.5)
        ):
        """
        A GradientCircle mask is a mask where the values ranges from 0 to 1. All pixels in the inner circle are set to 0,
        all pixels out of the outer cirlce are set to 1, and pixels in between have an intermediate value.
        
        Params:
        ---
        - inner_radius: int | float, the radius of the inner circle. If a float between 0 and 1 is provided, it refers to a fraction of min(width, height).
        For example, if the mask is loaded with a size of (100, 150) and inner_radius = 0.3, the inner circle has a radius of 30.
        If an int is provided, use it as the inner radius.
        - outer_radius: int | float, the radius of the outer circle. If a float between 0 and 1 is provided, it refers to a fraction of min(width, height).
        For example, if the mask is loaded with a size of (100, 150) and outer_radius = 0.4, the outer circle has a radius of 40.
        If an int is provided, use it as the outer radius. Default value is 0.5
        - transition: ZOZOCallable, a function mapping [0, 1] -> [0, 1], with transition(0) = 0 and transition(1) = 1. This function is used to apply
        non-linear gradients. It must be vectorized have only on param.
        - center: tuple[int, int] | tuple[float, float]. If two floats between 0 and 1 are provided, they refer to a fraction of width and height.
        For example, if the mask is loaded with a size of (100, 150) and a center of 0.3, 0.5, the mask contains circles centered in 30, 75.
        If ints are provided, they are directly used. The two values are independant, an int can be provided for the x component and a float for
        the second component (and vice versa).
        """
        super().__init__()
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        if not verify_ZOZOCallable(transition, test_vectorizaiton=True):
            raise ValueError("The provided transition isn't a ZOZOCallable.")
        self.transition = transition
        self.center = center

    def _load(self, width:int, height: int, **ld_kwargs):
        grid_x, grid_y = np.ogrid[:width, :height]

        center = (
            self.center[0]*width if 0 <= self.center[0] <= 1 else self.center[0],
            self.center[1]*height if 0 <= self.center[1] <= 1 else self.center[1]
        )
        distances = np.sqrt((grid_x - center[0] - 0.5) ** 2 + (grid_y - center[1] - 0.5) ** 2)
        inner_radius = self.inner_radius*min(height, width) if  0 <= self.inner_radius <= 1 else self.inner_radius
        outer_radius = self.outer_radius*min(height, width) if  0 <= self.outer_radius <= 1 else self.outer_radius
        distances = np.sqrt((grid_x - center[0] - 0.5) ** 2 + (grid_y - center[1] - 0.5) ** 2)
        self.matrix = np.clip((distances - inner_radius)/(outer_radius - inner_radius), 0, 1)
        self.matrix = self.transition(self.matrix)

class GradientRectangle(Mask):
    """
    A GradientRectangle mask is a mask where values range from 0 to 1. All pixels inside the inner rectangle are set to 0.
    All pixels outside the outer rectangle are set to 1. All pixels in between have an intermediate value.

    The intermediate value is defined by the transition function.
    """

    def __init__(
        self,
        inner_left: int,
        inner_right: int,
        inner_top: int,
        inner_bottom: int,
        outer_left: int = 0.,
        outer_right: int = 1.,
        outer_top: int = 0.,
        outer_bottom: int = 1.,
        transition: ZOZOCallable = linear
    ):
        """
        A GradientRectangle mask is a mask where values range from 0 to 1. All pixels inside the inner rectangle are set to 0.
        All pixels outside the outer rectangle are set to 1. All pixels in between have an intermediate value.

        Params:
        ---
        - inner_left: int | float, the coordinate of the left of the inner rectangle, included.
        - inner_top: int | float, the coordinate of the top of the inner rectangle, included.
        - inner_right: int | float, the coordinate of the right of the inner rectangle, included.
        - inner_bottom: int | float, the coordinate of the bottom of the inner rectangle, included.
        - outer_left: int | float, the coordinate of the left of the outer rectangle, included.
        - outer_top: int | float, the coordinate of the top of the outer rectangle, included.
        - outer_right: int | float, the coordinate of the right of the outer rectangle, included.
        - outer_bottom: int | float, the coordinate of the bottom of the outer rectangle, included.
        - transition: ZOZOCallable, a function mapping [0, 1] -> [0, 1], with transition(0) = 0 and transition(1) = 1. This function is used to apply
        non-linear gradients. It must be vectorized have only on param.
        """

        super().__init__()

        if outer_bottom < inner_bottom or outer_top > inner_top or outer_left > inner_left or outer_right < inner_right:
            raise ValueError(
                f"""The outer rectangle cannot be inside of the inner rectangle, got
                inner = ({inner_left, inner_right, inner_top, inner_bottom})
                and outer = ({outer_left, outer_right, outer_top, outer_bottom})"""
            )

        self.inner_left = inner_left
        self.inner_right = inner_right
        self.inner_bottom = inner_bottom
        self.inner_top = inner_top

        self.outer_left = outer_left
        self.outer_right = outer_right
        self.outer_bottom = outer_bottom
        self.outer_top = outer_top

        verify_ZOZOCallable(transition, test_vectorizaiton=True)
        self.transition = transition

    def _load(self, width: int, height: int, **ld_kwargs):
        y_indices, x_indices = np.meshgrid(np.arange(height), np.arange(width), indexing='ij')

        inner_left = self.inner_left*width if 0 <= self.inner_left <= 1 and isinstance(self.inner_left, float) else self.inner_left
        inner_right = self.inner_right*width if 0 <= self.inner_right <= 1 and isinstance(self.inner_right, float) else self.inner_right
        inner_top = self.inner_top*height if 0 <= self.inner_top <= 1 and isinstance(self.inner_top, float) else self.inner_top
        inner_bottom = self.inner_bottom*height if 0 <= self.inner_bottom <= 1 and isinstance(self.inner_bottom, float) else self.inner_bottom

        outer_left = self.outer_left*width if 0 <= self.outer_left <= 1 and isinstance(self.outer_left, float) else self.outer_left
        outer_right = self.outer_right*width if 0 <= self.outer_right <= 1 and isinstance(self.outer_right, float) else self.outer_right
        outer_top = self.outer_top*height if 0 <= self.outer_top <= 1 and isinstance(self.outer_top, float) else self.outer_top
        outer_bottom = self.outer_bottom*height if 0 <= self.outer_bottom <= 1 and isinstance(self.outer_bottom, float) else self.outer_bottom

        left_dist = np.clip((inner_left - x_indices) / (inner_left - outer_left + 1), 0, 1)
        right_dist = np.clip((x_indices - inner_right) / (outer_right - inner_right + 1), 0, 1)
        top_dist = np.clip((inner_top - y_indices) / (inner_top - outer_top + 1), 0, 1)
        bottom_dist = np.clip((y_indices - inner_bottom) / (outer_bottom - inner_bottom + 1), 0, 1)

        self.matrix = self.transition(np.clip(np.sqrt(left_dist**2 + right_dist**2 + top_dist**2 + bottom_dist**2), 0, 1))



"""The art module contains all the available arts for your game. You should look to arts based on geometries or on files."""
from .art import Art
from .file import ImageFile, ImageFolder, GIFFile
from .geometry import Polygon, Rectangle, RoundedRectangle, Circle, Ellipse, TexturedCircle, TexturedEllipse, TexturedPolygon, TexturedRoundedRectangle



"""The drawing module contains the drawing transformations."""
from typing import Sequence
from pygame import Surface, Rect
from pygamecv import rectangle, line, lines, polygon, circle, ellipse, pie, arc, rounded_rectangle
from .transformation import Transformation
from .._common import ColorValue

class DrawCircle(Transformation):
    """Draw a circle on the art."""

    def __init__(
        self,
        color: ColorValue,
        radius: int,
        center: tuple[int, int],
        thickness: int = 0,
        allow_antialias: bool = True
    ) -> None:
        """
        Draw a circle on the art.

        Params:
        ----
        - color: ColorValue, the color used to draw the circle. It can have an alpha channel != 255.
        - radius: int, the radius of the circle
        - center: tuple[int, int], the center of the circle
        - thickness: int, the thickness of the line used to draw the circle. If 0, the circle is filled.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        """
        super().__init__()

        self.radius = radius
        self.color = color
        self.thickness = thickness
        self.center = center
        self.allow_antialias = allow_antialias

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        antialias = self.allow_antialias and ld_kwargs.get("antialias", False)
        for surf in surfaces:
            circle(surf, self.center, self.radius, self.color, self.thickness, antialias)
        return surfaces, durations, introduction, None, width, height
    
    def cost(self, width: int, height: int, length: int, **ld_kwargs):
        if self.color.a == 255 and not ld_kwargs.get("antialias", False):
            return self.radius**2*4*length
        else:
            return self.radius**2*4*length*4 # Once for the color draw, twice for the apha rendering (copy and addWeighted) and once for the final blit.

class DrawRectangle(Transformation):
    """Draw a rectangle on the art."""
    def __init__(
        self,
        color: ColorValue,
        rect: Rect,
        thickness: int,
        allow_antialias: bool = True
    ) -> None:
        """
        Draw a rectangle on the art.

        Params:
        ---
        - color: ColorValue, the color used to draw the rectangle. It can have an alpha channel != 255.
        - rect: pygame.Rect, the rectangle representing the edges of the rectangle.
        - thickness: int, the thickness of the line used to draw the rectangle. If 0, the rectangle is filled.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        """

        super().__init__()  
        self.color = color
        self.rect = Rect(rect)
        self.allow_antialias = allow_antialias
        self.thickness = thickness

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        for surf in surfaces:
            rectangle(surf, self.rect, self.color, self.thickness)
        return surfaces, durations, introduction, None, width, height

    def cost(self, width: int, height: int, length: int, **ld_kwargs):
        if self.color.a == 255 and not ld_kwargs.get("antialias", False):
            return self.rect.width*self.rect.height*length
        else:
            return self.rect.width*self.rect.height*length*4 # Once for the color draw, twice for the apha rendering (copy and addWeighted) and once for the final blit.

class DrawRoundedRectantle(Transformation):
    """Draw a rectangle on the art, with rounded corners."""
    def __init__(
        self,
        color: ColorValue,
        rect: Rect,
        top_left: int,
        top_right: int = None,
        bottom_right: int = None,
        bottom_left: int = None,
        thickness: int = 0,
        allow_antialias: bool = True
    ) -> None:
        """
        Draw a rectangle on the art, with rounded corners.

        Params:
        ---
        - color: ColorValue, the color used to draw the rounded rectangle. It can have an alpha channel != 255.
        - rect: pygame.Rect, the rectangle representing the edges of the rectangle.
        - top_left, top_right, bottom_left, bottom_right: the radii of the rounded corners.
        If one of the 3 last is None, the value for the top_left corner is used instead.
        - thickness: int, the thickness of the line used to draw the rounded rectangle. If 0, the rounded rectangle is filled.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        """
        super().__init__()  
        self.color = color
        self.rect = Rect(rect)
        self.thickness = thickness
        self.allow_antialias = allow_antialias
        self.top_left = top_left
        self.top_right = top_right
        self.bottom_right = bottom_right
        self.bottom_left = bottom_left

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        antialias = self.allow_antialias and ld_kwargs.get("antialias", False)
        for surf in surfaces:
            rounded_rectangle(surf, self.rect, self.color, self.thickness, antialias, self.top_left, self.top_right, self.bottom_left, self.bottom_right)
        return surfaces, durations, introduction, None, width, height

    def cost(self, width: int, height: int, length: int, **ld_kwargs):
        if self.color.a == 255 and (not ld_kwargs.get("antialias", False) or self.top_right == self.top_left == self.bottom_right == self.bottom_left == 0) and (
        # Condition to pygame.draw in pygame.CV
            self.top_right + self.top_left <= self.rect.width//2
            and self.bottom_left + self.bottom_right <= self.rect.width//2
            and self.top_right + self.bottom_right <= self.rect.height//2
            and self.top_left + self.bottom_left <= self.rect.height//2
        ):
            return self.rect.width*self.rect.height*length
        else:
            return self.rect.width*self.rect.height*length*4 # Once for the color draw, twice for the apha rendering (copy and addWeighted) and once for the final blit.

class DrawEllipse(Transformation):
    """Draw an ellipse on the art."""

    def __init__(
        self,
        color: ColorValue,
        radius_x: int,
        radius_y: int,
        center: tuple[int, int],
        thickness: int = 0,
        angle: int=0,
        allow_antialias: bool = True
    ):
        """
        Draw an ellipse on the art.

        Params:
        ----
        - color: ColorValue, the color used to draw the ellipse. It can have an alpha channel != 255.
        - radius_x: int, the horizontal radius of the ellipse, before rotation.
        - radius_y: int, the vertical radius of the ellipse, before rotation.
        - center: tuple[int, int], the center of the ellipse
        - thickness: int, the thickness of the line used to draw the ellipse. If 0, the ellipse is filled.
        - angle: int = 0, the angle in degrees by which the ellipse should be rotated, counterclockwise.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        """
        self.color = color
        self.radius_x = radius_x
        self.radius_y = radius_y
        self.center = center
        self.angle = angle
        self.thickness = thickness
        self.allow_antialias = allow_antialias

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        antialias = self.allow_antialias and ld_kwargs.get("antialias", False)
        for surf in surfaces:
            ellipse(surf, self.center, self.radius_x, self.radius_y, self.color, self.thickness, antialias, self.angle)
        return surfaces, durations, introduction, None, width, height

    def cost(self, width: int, height: int, length: int, **ld_kwargs):
        if self.color.a == 255 and not ld_kwargs.get("antialias", False):
            return self.radius_x*self.radius_y*4*length
        else:
            return self.radius_x*self.radius_y*4*length*4 # Once for the color draw, twice for the apha rendering (copy and addWeighted) and once for the final blit.

class DrawPolygon(Transformation):
    """Draw a polygon on the art."""

    def __init__(
        self,
        color: ColorValue,
        points: Sequence[tuple[int, int]],
        thickness: int = 0,
        allow_antialias: bool = True
    ) -> None:
        """
        Draw a polygon on the art.

        Params:
        ---
        - color: ColorValue, the color used to draw the polygon. It can have an alpha channel != 255.
        - points: Sequence[tuple[int, int]], the sequence of points.
        - thickness: int, the thickness of the line used to draw the polygon. If 0, the polygon is filled.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        """
        super().__init__()

        self.color = color
        self.points = points
        self.thickness = thickness
        self.allow_antialias = allow_antialias

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        antialias = self.allow_antialias and ld_kwargs.get("antialias", False)
        for surf in surfaces:
            polygon(surf, self.points, self.color, self.thickness, antialias)
        return surfaces, durations, introduction, None, width, height

    def cost(self, width: int, height: int, length: int, **ld_kwargs):
        left = min(point[0] for point in self.points) - self.thickness//2
        right = max(point[0] for point in self.points) + self.thickness//2 +1
        top = min(point[1] for point in self.points) - self.thickness//2
        bottom = max(point[1] for point in self.points) + self.thickness//2 + 1
        width = right - left +1
        height = bottom - top +1
        return width*height*length*4 # Once for the color draw, twice for the apha rendering (copy and addWeighted) and once for the final blit.

class DrawLine(Transformation):
    """Draw one line on the art."""

    def __init__(self, color: ColorValue, p1: tuple[int, int], p2: tuple[int, int], thickness: int = 1, allow_antialias: bool = True) -> None:
        """
        Draw one line on the art.
        
        Params:
        ---
        - color: ColorValue, the color used to draw the line. It can have an alpha channel != 255.
        - p1: tuple[int, int], the position of start of the line.
        - p2: tuple[int, int], the position of the end of the line.
        - thickness: int, the thickness of the line used to draw the line.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        """

        self.color = color
        self.p1 = p1
        self.p2 = p2
        self.thickness = thickness
        self.allow_antialias = allow_antialias
        super().__init__()

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        antialias = self.allow_antialias and ld_kwargs.get("antialias", False)
        for surf in surfaces:
            line(surf, self.p1, self.p2, self.color, self.thickness, antialias)
        return surfaces, durations, introduction, None, width, height

    def cost(self, width: int, height: int, length: int, **ld_kwargs):
        left = min(self.p1[0], self.p2[0]) - self.thickness//2
        right = max(self.p1[0], self.p2[0]) + self.thickness//2 
        top = min(self.p1[1], self.p2[1]) - self.thickness//2
        bottom = max(self.p1[1], self.p2[1]) + self.thickness//2
        width = right - left + 1
        height = bottom - top + 1
        return width*height*length*4 # Once for the color draw, twice for the apha rendering (copy and addWeighted) and once for the final blit.

class DrawLines(Transformation):
    """Draw lines on the art."""

    def __init__(self, color: ColorValue, points: Sequence[tuple[int, int]], thickness: int = 1, closed: bool = False, allow_antialias: bool = True) -> None:
        """
        Draw lines on the art. This is faster than drawing the lines one by one.

        Params:
        ---
        - color: ColorValue, the color used to draw the lines. It can have an alpha channel != 255.
        - points: Sequence[tuple[int, int]], the sequence of points.
        - thickness: int, the thickness of the line used to draw the lines.
        - closed: bool, If True, the last points is linked to the first point with a line.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        """
        self.color = color
        self.points = points
        self.thickness = thickness
        self.closed = closed
        self.allow_antialias = allow_antialias
        super().__init__()

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        antialias = self.allow_antialias and ld_kwargs.get("antialias", False)
        for surf in surfaces:
            lines(surf, self.points, self.color, self.thickness, antialias, self.closed)
        return surfaces, durations, introduction, None, width, height

    def cost(self, width: int, height: int, length: int, **ld_kwargs):
        left = min(point[0] for point in self.points) - self.thickness//2
        right = max(point[0] for point in self.points) + self.thickness//2 +1
        top = min(point[1] for point in self.points) - self.thickness//2
        bottom = max(point[1] for point in self.points) + self.thickness//2 + 1
        width = right - left +1
        height = bottom - top +1
        return width*height*length*4 # Once for the color draw, twice for the apha rendering (copy and addWeighted) and once for the final blit.

class DrawArc(Transformation):
    """Draw an arc on the art."""

    def __init__(
        self,
        color: ColorValue,
        radius_x: int,
        radius_y: int,
        ellipsis_center: tuple[int, int],
        from_angle: int,
        to_angle: int,
        thickness: int = 1,
        angle: int = 0,
        allow_antialias: bool = True
    ) -> None:
        """
        Draw an arc on the art. An arc is a part of an ellipse.

        Params:
        ----
        - color: ColorValue, the color used to draw the arc. It can have an alpha channel != 255.
        - radius_x: int, the horizontal radius of the ellipse, before rotation.
        - radius_y: int, the vertical radius of the ellipse, before rotation.
        - ellipsis_center: tuple[int, int], the center of the ellipse.
        - from_angle: the starting angle of the arc, in degrees
        - to_angle: the ending angle of the arc, in degrees. The arc is drawn counterclockwise.
        - thickness: int, the thickness of the line used to draw the arc. If 0, a filled pie is drawn.
        - angle: int = 0, the angle in degrees by which the ellipse should be rotated, counterclockwise.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        """
        self.color = color
        self.thickness = thickness
        self.ellipsis_center = ellipsis_center
        self.rx = radius_x
        self.ry = radius_y
        self.from_angle = from_angle
        self.to_angle = to_angle
        self.angle = angle
        self.allow_antialias = allow_antialias
        super().__init__()

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        antialias = self.allow_antialias and ld_kwargs.get("antialias", False)
        for surf in surfaces:
            arc(surf, self.ellipsis_center, self.rx, self.ry, self.color, self.thickness, antialias, self.angle, self.from_angle, self.to_angle)

        return surfaces, durations, introduction, None, width, height

    def cost(self, width: int, height: int, length: int, **ld_kwargs):
        return self.rx*self.ry*4*length*4 # Once for the color draw, twice for the apha rendering (copy and addWeighted) and once for the final blit.

class DrawPie(Transformation):
    """Draw a pie on the art."""

    def __init__(
        self,
        color: ColorValue,
        radius_x: int,
        radius_y: int,
        ellipsis_center: tuple[int, int],
        from_angle: int,
        to_angle: int,
        thickness: int = 1,
        angle: int = 0,
        allow_antialias: bool = True
    ) -> None:
        """
        Draw a pie on the art. A pie is an arc where the start and end points are linked to the center of the ellipse.

        Params:
        ----
        - color: ColorValue, the color used to draw the pie. It can have an alpha channel != 255.
        - radius_x: int, the horizontal radius of the ellipse, before rotation.
        - radius_y: int, the vertical radius of the ellipse, before rotation.
        - ellipsis_center: tuple[int, int], the center of the ellipse.
        - from_angle: the starting angle of the arc, in degrees
        - to_angle: the ending angle of the arc, in degrees. The arc is drawn counterclockwise.
        - thickness: int, the thickness of the line used to draw the pie. If 0, the pie is filled.
        - angle: int = 0, the angle in degrees by which the ellipse should be rotated, counterclockwise.
        - allow_antialias: bool = True, if False, even with 'antialias' : True as a loading kwarg, the drawing will be done without antialias.
        """

        self.color = color
        self.thickness = thickness
        self.ellipsis_center = ellipsis_center
        self.rx = radius_x
        self.ry = radius_y
        self.from_angle = from_angle
        self.to_angle = to_angle
        self.angle = angle
        self.allow_antialias = allow_antialias
        super().__init__()

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        antialias = self.allow_antialias and ld_kwargs.get("antialias", False)
        for surf in surfaces:
            pie(surf, self.ellipsis_center, self.rx, self.ry, self.color, self.thickness, antialias, self.angle, self.from_angle, self.to_angle)
        return surfaces, durations, introduction, None, width, height

    def cost(self, width: int, height: int, length: int, **ld_kwargs):
        return self.rx*self.ry*4*length*4 # Once for the color draw, twice for the apha rendering (copy and addWeighted) and once for the final blit.



"""The art class is the base for all the surfaces and animated surfaces of the game."""
from abc import ABC, abstractmethod
from threading import Thread
from pygame import Surface, image, surfarray as sa, Rect
from PIL import Image
from ..transform import Transformation, Pipeline, ExtractSlice, ExtractOne
from .._common import LoadingError

class Art(ABC):
    """The Art class is the base for all the surfaces and animated surfaces of the game. They cannot be instanciated."""

    def __init__(self, transformation: Transformation = None) -> None:
        """
        Creates an Art.

        Params:
        ----
        - transformation: transform.Transformation = None. Any transformation (or Pipeline) that will be applied to the art when it is loaded.
        """
        super().__init__()
        self._surfaces: tuple[Surface] = ()
        self._durations: tuple[int] = ()
        self._introduction = 0
        self._loaded = False

        self._time_since_last_change = 0
        self._index = 0

        self._height = -1
        self._width = -1
        self._on_loading_transformation = transformation

        self._buffer_transfo_pipeline = Pipeline()

        self._transfo_thread = None
        self._has_changed = False
        self._copies: list[_ArtAsCopy] = []
        self._references: list[_ArtAsReference] = []

    @property
    def surfaces(self):
        """Return the surfaces of the art in order of display."""
        return self._surfaces

    @property
    def durations(self):
        """Return the durations of the frames in order of display."""
        return self._durations

    @property
    def introduction(self):
        """Return the introduction index of the art."""
        return self._introduction

    def _find_initial_dimension(self):
        """Find the initial dimension of the art. To be called at the end of initialization."""
        if self._on_loading_transformation:
            self._width, self._height = self._on_loading_transformation.get_new_dimension(self._width, self._height)

    def _verify_sizes(self):
        """verify that all surfaces have the same sizes."""
        heights = set(surf.get_height() for surf in self.surfaces)
        widths = set(surf.get_width() for surf in self.surfaces)
        if len(heights) != 1:
            raise LoadingError(f"All images of the art does not have the same height, got\n{heights}")
        if len(widths) != 1:
            raise LoadingError(f"All images of the art does not have the same width, got\n{widths}")

    @property
    def size(self):
        """Return the size of the art."""
        return (self.width, self._height)

    @property
    def height(self):
        """Return the height of the art."""
        return self._height

    @property
    def width(self):
        """Return the width of the art."""
        return self._width

    def is_loaded(self):
        """Return true if the art is loaded"""
        return self._loaded

    @property
    def total_duration(self):
        """Return the durations of the frames in the art."""
        if len(self.durations) > 1:
            return sum(self.durations)
        return 0

    @abstractmethod
    def _load(self, **ld_kwargs):
        """
        Sublcass-specific loading method of the art.
        """
        raise NotImplementedError()

    @property
    def index(self):
        """Return the current index of the frame displayed."""
        return self._index

    def unload(self):
        """Unload the surfaces and reset."""
        self.reset() # Reset the index.
        if self._transfo_thread is not None:
            # Wait for the transformation thread to stop to not get surfaces still loaded in memory.
            self._transfo_thread.join()
        self._surfaces = ()
        self._durations = ()
        self._loaded = False

    def load(self, **ld_kwargs):
        """Load the art and all its copies. If the Art is already loaded, only the copies not loaded are loaded. 
        
        Params:
        ----
        - **ld_kwargs: The loading keyword arguments. the arguments needed for the arts to be loaded. Should include 'antialias' and 'cost_threshold',
        if they are not provided, default values are used: False for antialias and 200_000 for the cost. Other entries can be given if some custom arts
        need them.
        """

        if not self._loaded:
            self.reset()
            self._load(**ld_kwargs)
            self._verify_sizes()
            self._loaded = True
            if not self._on_loading_transformation is None:
                self._transform(self._on_loading_transformation, **ld_kwargs)

        for copy in self._copies:
            copy.load(**ld_kwargs)

    def update(self, loop_duration: float) -> bool:
        """
        Update the instance animation.
        
        Params:
        ---
        - loop_duration: float, the duration of the game loop, (the value returned by clock.tick(FPS)).

        Returns:
        ---
        - has_changed: bool, whether the index changed or a transformation have been applied since the last call.
        """
        if len(self.surfaces) > 1:
            self._time_since_last_change += loop_duration
            if self._time_since_last_change >= self.durations[self._index]:
                self._time_since_last_change -= self.durations[self._index]
                self._index += 1
                # Loop
                if self._index == len(self.surfaces):
                    self._index = self.introduction
                # If the loop_duration is too long, we reset the time to 0.
                if self._time_since_last_change >= self.durations[self._index]:
                    self._time_since_last_change = 0

                self._has_changed = True
        has_changed = self._has_changed # This can be set to True if the a transformation has been a applied recently, or if the index changed.
        self._has_changed = False
        return has_changed 

    def reset(self):
        """Reset the animation."""
        self._index = 0
        self._time_since_last_change = 0

    def get(self, match: 'Art' = None, **ld_kwargs) -> Surface:
        """
        Return the current Frame. If not already loaded, the art is loaded now. If a transformation is waiting to be executed, it is executed now.
        
        Params:
        ---
        - match: Art, if povided, the index will match the index of the other art to match, otherwise, use its own index.
        - **ld_kwargs: The loading keyword arguments. the arguments needed for the arts to be loaded. Should include 'antialias' and 'cost_threshold',
        if they are not provided, default values are used: False for antialias and 200_000 for the cost. Other entries can be given if some custom arts
        need them.
        """
        if not self._loaded: # Load the art
            self.load(**ld_kwargs)

        if (
            not self._buffer_transfo_pipeline.is_empty()
            and (self._transfo_thread is None or not self._transfo_thread.is_alive())
        ): # Apply a transformation only if the last thread is finished
            if self._buffer_transfo_pipeline.cost(self.width, self.height, len(self), **ld_kwargs) < ld_kwargs.get("cost_threshold", 200_000):
                args = self._buffer_transfo_pipeline.copy(), # On a separate thread, in this case the transformation may be visible later.
                self._buffer_transfo_pipeline.clear()
                self._transfo_thread = Thread(target=self._transform, args=args, kwargs=ld_kwargs)
                self._transfo_thread.start()
            else:
                self._transform(self._buffer_transfo_pipeline, **ld_kwargs) # Or directly on the main thread.
                self._buffer_transfo_pipeline.clear()

        index = self._index if match is None else match.index
        return self.surfaces[index] # The current surface
    
    def transform(self, transformation: Transformation):
        """Apply a transformation to an Art.
        
        Params:
        ----
        - transformation: Transformation, the transformation that will be applied in the next get.
        """
        self._buffer_transfo_pipeline.add_transformation(transformation)

    def _transform(self, transformation: Transformation, **ld_kwargs):
        """
        Apply a transformation.
        
        Params:
        ----
        - transformation: Transformation, the transformation that will be applied in the next get.
        - **ld_kwargs: The loading keyword arguments. the arguments needed for the arts to be loaded. Should include 'antialias' and 'cost_threshold',
        if they are not provided, default values are used: False for antialias and 200_000 for the cost. Other entries can be given if some custom arts
        need them.
        """
        (   self._surfaces,
            self._durations,
            self._introduction,
            index,
            self._width,
            self._height
        ) = transformation.apply(
            self._surfaces,
            self._durations,
            self._introduction,
            self._index,
            self._width,
            self._height,
            **ld_kwargs
        )
        if index is not None:
            self._index = index
        self._buffer_transfo_pipeline.clear()
        self._has_changed = True
        for reference in self._references:
            reference._has_changed = True

    def copy(self, additional_transformation: Transformation = None) -> '_ArtAsCopy':
        """
        Return an independant copy of the art. The copy's initial surfaces will be the same as its original. If a copy is created after the loading
        of the original, its initial surfaces is the original's current surfaces and not the original's initial surfaces.

        Params:
        ---
        - additional_transformation: Transformation = None. If provided, another transformation will be performed on the copy during its loading.
        """
        copy = _ArtAsCopy(self, additional_transformation)
        self._copies.append(copy)
        return copy
    
    def reference(self) -> '_ArtAsReference':
        """
        Return a dependant copy of the art. This new object shares the surfaces and durations, any transformation applied to it will be applied
        to the original art (and vice versa). The animation, however, is independant.
        """
        reference = _ArtAsReference(self)
        self._references.append(reference)
        return reference

    def get_rect(self) -> Rect:
        """Create a pygame.Rect without masked based on this art."""
        return self.surfaces[0].get_rect()

    def save(self, path: str, index: int | slice = None):
        """
        Save the art as a gif or as an image.
        
        Params:
        ----
        - path: str, the path where the art should be saved.
        - index: int | slice = None, the index or indices of the frames in the animation that will be saved.
        
        """
        if not self.is_loaded():
            raise LoadingError("Cannot save an unloaded art.")
        if len(self.surfaces) == 1:
            image.save(self.surfaces[0], path)
        elif index is not None and isinstance(index, int):
            image.save(self.surfaces[index%len(self.surfaces)], path)
        else:
            surfaces = self.surfaces[index] if isinstance(index, slice) else self.surfaces
            durations = self.durations[index] if isinstance(index, slice) else self.durations
            pil_images = [Image.fromarray(sa.array3d(surf)) for surf in surfaces]
            pil_images[0].save(path, format='GIF', save_all=True, append_images = pil_images[1:], duration=durations)

    def __len__(self):
        return len(self._surfaces)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self.copy(ExtractSlice(key))
        elif isinstance(key, int):
            return self.copy(ExtractOne(key))
        else:
            raise IndexError(f"Art indices must be integers or slices, not {type(key)}")

class _ArtAsCopy(Art):
    """ArtAsCopy represent an Art created with the .copy() method of another art."""

    def __init__(self, original: Art, additional_transformation: Transformation):
        super().__init__(additional_transformation)
        # The on load transformation has been removed because the transformation are executed during the loading of the original
        self._original = original
        self._height = self._original.height
        self._width = self._original.width
        self._find_initial_dimension()

    def _load(self, **ld_kwargs):
        if not self._original.is_loaded():
            self._original.load(**ld_kwargs)

        self._surfaces = tuple(surf.copy() for surf in self._original.surfaces)
        self._durations = self._original.durations
        self._introduction = self._original.introduction

class _ArtAsReference(Art):
    """ArtAsReference represent an Art created with the .reference() method of another art."""

    def __init__(self, original: Art):

        super().__init__(None)
        self._original = original
        self._height = self._original.height
        self._width = self._original.width

    def _load(self, **ld_kwargs):
        if not self._original.is_loaded():
            self._original.load(**ld_kwargs)

    def _transform(self, transformation: Transformation, **ld_kwargs):
        self._original._transform(transformation, **ld_kwargs)

    @property
    def surfaces(self):
        return self._original.surfaces

    @property
    def durations(self):
        return self._original.durations

    @property
    def introduction(self):
        return self._original.introduction



"""The commone module contains the LoadingError Exception and some common objects."""
from typing import Union, Tuple, Sequence
from pygame import Color

class LoadingError(Exception):
    """Error to be raised when an error related to the loading of an Art occurs."""

ColorValue = Union[Color, int, str, Tuple[int, int, int], Tuple[int, int, int, int], Sequence[int]]



"""
gamarts is a python library used to represent animations and static images with the same Art class.
It also introduce clever loading and unloading methods.
"""
from gamarts.art import (
    GIFFile, ImageFile, ImageFolder, Rectangle, RoundedRectangle, Circle, Ellipse, Polygon,
    TexturedCircle, TexturedEllipse, TexturedPolygon, TexturedRoundedRectangle, Art
)
import gamarts.mask as mask
import gamarts.transform as transform

LD_KWARGS = {'antialias': False, 'cost_threshold': 200_000}



"""The binary transformation file contains transformation that include other arts."""
from pygame import Surface, transform as tf
from .transformation import Transformation

class Concatenate(Transformation):
    """The concatenate transformation concatenates multiple arts into one bigger animation."""

    def __init__(self, *others) -> None:
        """
        The concatenate transformation concatenates multiple arts into a longer one animation.

        Params:
        ---
        - *others: Art, the other arts the transformed art will be concatenated with.

        Raises:
        ---
        - ValueError, if all the arts do not have the exact same size.
        """
        super().__init__()
        if len(set(other.size for other in others)) != 1:
            raise ValueError("All arts must have the same size to be concatenated.")
        self.others = others

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        need_to_unloads = []
        for art in self.others:
            if not art.is_loaded():
                need_to_unloads.append(True)
                art.load(**ld_kwargs)
            else:
                need_to_unloads.append(False)

        surfaces = surfaces + sum((art.surfaces for art in self.others), ())
        durations = durations + sum((art.durations for art in self.others), ())

        for other, need_to_unload in zip(self.others, need_to_unloads):
            if need_to_unload:
                other.unload()

        return surfaces, durations, introduction, None, width, height

def _combine_arts(*durations: tuple[int], introduction: int) -> tuple[list[tuple[int, tuple[int]]], int]:
    """Combine a list of durations to create a new art."""
    indexes = [0 for _ in durations]
    output = []
    tot_time = 0
    mx = max(sum(d) for d in durations)
    matrix_duration = [list(d) for d in durations]

    output_introduction = None
    for durations in matrix_duration:
        durations[-1] = durations[-1] + (mx - sum(durations))

    while tot_time < mx:

        if indexes[0] == introduction and output_introduction is None:
            output_introduction = len(output)

        mn = min(dur[0] for dur in matrix_duration)
        output.append((mn, tuple(indexes)))

        for i, durations in enumerate(matrix_duration):
            durations[0] -= mn
            if durations[0] == 0:
                durations.pop(0)
                indexes[i] += 1
        tot_time += mn

    return output, output_introduction

class Average(Transformation):
    """
    Compute the average of the frames of the art. The average is computed taking into account the time and durations of the frames.
    If the art have different total durations, the last frame of shorter arts is extend.
    """

    def __init__(self, *others) -> None:
        """
        Compute the average of the frames of the art.

        Params:
        ---
        - *others: Art, the other arts the transformed art will be averaged with.

        Raises:
        ---
        - ValueError, if all the arts do not have the exact same size.
        """
        super().__init__()
        if len(set(other.size for other in others)) != 1:
            raise ValueError("All arts must have the same size to be concatenated.")
        self.others = others

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        need_to_unloads = []
        for art in self.others:
            if not art.is_loaded():
                need_to_unloads.append(True)
                art.load(**ld_kwargs)
            else:
                need_to_unloads.append(False)

        combined_arts, introduction = _combine_arts(durations, *(other.durations for other in self.others), introduction=introduction)
        new_surfaces = []
        new_durations = []
        all_surfaces = [surfaces] + [other.surfaces for other in self.others]
        for duration, (indexes) in combined_arts:
            avg = tf.average_surfaces([surf[index] for surf, index in zip(all_surfaces, indexes)])
            new_surfaces.append(avg)
            new_durations.append(duration)

        for other, need_to_unload in zip(self.others, need_to_unloads):
            if need_to_unload:
                other.unload()

        return new_surfaces, new_durations, introduction, None, width, height

class Blit(Transformation):
    """
    Copy an art over another one. The blitting is computed taking into account the time and durations of the frames.
    If the art have different total durations, the last frame of shortest art is extend.
    """

    def __init__(self, other, x: int, y: int) -> None:
        """
        Copy an art over another one.
        
        Params:
        ---
        - other: art, the art to be blitted on top of this one.
        - x, y: int, the coordinate on which the other art is blitted.
        """
        super().__init__()
        self.other = other
        self.pos = (x,y)

    def apply(self, surfaces: tuple[Surface], durations: tuple[int], introduction: int, index: int, width: int, height: int, **ld_kwargs):
        need_to_unload = False
        if not self.other.is_loaded():
            self.other.load(**ld_kwargs)
            need_to_unload = True

        combined_arts, introduction = _combine_arts(durations, self.other.durations, introduction=introduction)
        new_surfaces = []
        new_durations = []
        for duration, (idx1, idx2) in combined_arts:
            back = surfaces[idx1].copy()
            back.blit(self.other.surfaces[idx2], self.pos)
            new_surfaces.append(back)
            new_durations.append(duration)

        if need_to_unload:
            self.other.unload()

        return new_surfaces, new_durations, introduction, None, width, height



"""
The mask module contains masks. Masks are objects used to select part of an image on which apply an effect.
There are 6 types of Masks:
The first type is composed of only one class: MatrixMask
The second one is composed of binary geometrical masks: Circle, Ellipsis, Rectangle etc.
The third one is composed of gradient geometrical masks: GradientCircle, GradientRectangle etc.
The fourth one is composed of masks extracted from arts or from images.
The fifth one is combinations or transformation of other masks.
The last one is moving masks.
"""
from .mask import Mask, MatrixMask, Circle, Ellipse, GradientCircle, GradientRectangle, Rectangle, RoundedRectangle, Polygon
from .transformation import (
    FromArtAlpha, FromArtColor, FromImageColor, SumOfMasks, ProductOfMasks, AverageOfMasks, DifferenceOfMasks,
    BinaryMask, BlitMaskOnMask, InvertedMask, TransformedMask
)



from setuptools import setup, find_packages

setup(
    name='gamarts',
    author="Tanguy Dugas du Villard",
    author_mail="tanguy.dugas01@gmail.com",
    version='1.0.0',
    description="Gamarts is a python library providing a unique way to represent static and animated surfaces in pygame, alongside with a clever loading and unloading behavior.",
    packages=find_packages(),
    install_requires=[
        'pygame',
        'pygame-cv',
        'ZOCallable',
        'pillow'
    ],
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Tanguy-ddv/pygame-arts",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Games/Entertainment"
    ],
    python_requires='>=3.6'
)